# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
import numpy as np
from six.moves import range


class BaseFilter(object):
    def filter(self, nparray_dict):
        raise NotImplementedError()

    def _do_filter(self, base_filter, nparray_dict):
        if hasattr(base_filter, 'shape'):
            # base_filter is boolean array result of filtering
            # over multidimensional array, for example:
            #      [[True, False, False, True],
            #       [True, False, True, False]]
            #
            # Combine the boolean arrays by 'or', for the above example
            # resulting in:
            #      [True, False, True, True]

            if len(base_filter.shape) == 2:
                # Combine filters with | (or)
                temp_filter = base_filter[0]
                for x in range(1, base_filter.shape[0]):
                    temp_filter = temp_filter | base_filter[x]

                base_filter = temp_filter

        for key in nparray_dict:
            # Apply the filter on the np_array's
            if nparray_dict[key] is not None and hasattr(
               nparray_dict[key], 'shape'):

                # Transform the base_filter by prepending slice(None) to
                # match the dimensionality of the nparray_dict[key].shape
                #
                #      shape(100,) => _filter = [base_filter]
                #      shape(2, 100)  => _filter = [slice(None), base_filter]
                #
                #      Note: x[slice(None),[1,2,3]] == x[:,[1,2,3]]

                _filter = [slice(None)] * (
                    len(nparray_dict[key].shape) - 1) + [base_filter]

                # try both with tuple and list
                # This solves h5py datasets indexing errors
                nparray_dict[key] = nparray_dict[key][:][tuple(_filter)]

    def filter_dict(self, nparray_dict, model):
        """
        param: nparray_dict = dictionairy of np_array's.
        """
        # Get the filter
        base_filter = self.filter(nparray_dict)

        self._do_filter(base_filter, nparray_dict)

    def get_boolean_mask_filter(self, nparray_dict, model):
        base_filter = self.filter(nparray_dict)
        self._do_filter(base_filter, nparray_dict)

        return base_filter

    def get_field_name(self):
        raise NotImplementedError()

    def to_dict(self):
        raise NotImplementedError()


class BaseCompareFilter(BaseFilter):
    """
    Compare filter base
    """
    func_str = '=='
    func_name = 'eq'

    def __init__(self, key, field, value, filter_as=False):
        """
        :param key: E.g. 'content_pk'
        :param field: GridDataField instance
        :param value: E.g. 3, [1, 2], etc.
        """
        self._key = key
        self._field = field
        self._value = value
        self._filter_as = filter_as

    def get_field_name(self):
        return self._key

    def __repr__(self):
        return "{0}({1} {2} {3})".format(
            self.__class__.__name__,
            self.func_str,
            self._key, self._value)

    def to_dict(self):
        return {
            'filter': {
                "{0}__{1}".format(
                    self._key, self.func_name): self._value
            }
        }


class EqualsFilter(BaseCompareFilter):
    func_str = '=='
    func_name = 'eq'

    def filter(self, nparray_dict):
        return nparray_dict[self._key][:] == self._value


class NotEqualsFilter(BaseCompareFilter):
    func_str = '!='
    func_name = 'ne'

    def filter(self, nparray_dict):
        return nparray_dict[self._key][:] != self._value


class GtFilter(BaseCompareFilter):
    func_str = '>'
    func_name = 'gt'

    def filter(self, nparray_dict):
        return nparray_dict[self._key][:] > self._value


class GteFilter(BaseCompareFilter):
    func_str = '>='
    func_name = 'gte'

    def filter(self, nparray_dict):
        return nparray_dict[self._key][:] >= self._value


class LtFilter(BaseCompareFilter):
    func_str = '<'
    func_name = 'lt'

    def filter(self, nparray_dict):
        return nparray_dict[self._key][:] < self._value


class LteFilter(BaseCompareFilter):
    func_str = '<='
    func_name = 'lte'

    def filter(self, nparray_dict):
        return nparray_dict[self._key][:] <= self._value


class InFilter(BaseCompareFilter):
    func_str = ' in '
    func_name = 'in'

    def __init__(self, key, field, value, filter_as):
        try:
            iter(value)
        except TypeError:
            raise Exception("Values should be iterable")

        self._field = field
        self._key = key
        self._value = value
        self._filter_as = filter_as

    def filter(self, nparray_dict):
        # Concatenate equals with | (or)
        return np.isin(nparray_dict[self._key][:], self._value)


class SliceFilter(BaseFilter):
    """
    Slice filter
    """
    def __init__(self, _slice):
        self._slice = _slice

    def filter(self, nparray_dict):
        return self._slice

    def get_field_name(self):
        return None

    def __repr__(self):
        return "SliceFilter({0})".format(
            self._slice)

    def to_dict(self):
        return {'slice': [
            self._slice.start, self._slice.stop, self._slice.step]}


FILTER_MAP = {
    'eq': EqualsFilter,
    'ne': NotEqualsFilter,
    'gt': GtFilter,
    'gte': GteFilter,
    'lt': LtFilter,
    'lte': LteFilter,
    'in': InFilter,
}


def get_filter(
        splitted_keys, field, value, filter_map=FILTER_MAP, filter_as=False):
    """
    Helper function for getting a filter.

    :param splitted_keys: list containing field_name and an optional
        filter_name
    :param field: an instance of GridDataField for the field_name
        (splitted_keys[0])
    :param value: value or values to filter

    Returns: an instantiated filter (a subclass of BaseFilter)

    Example:

    >>> from threedigrid.orm.base.fields import ArrayField
    >>> get_filter(['content_pk', 'in'], ArrayField(), [1, 2, 3, 4])
    InFilter( in  content_pk [1, 2, 3, 4])
    """

    # Default to FILTER_MAP
    if filter_map is None:
        filter_map = FILTER_MAP

    len_splitted = len(splitted_keys)
    if len_splitted == 1:
        field_name, filter_name = splitted_keys[0], 'eq'
    elif len_splitted == 2:
        field_name, filter_name = splitted_keys
    else:
        raise ValueError("splitted keys can only contain one or two elements")

    try:
        filter_class = filter_map[filter_name]
    except KeyError:
        raise ValueError("'%s' is an unknown filter name" % filter_name)

    return filter_class(field_name, field, value, filter_as)



# -*- coding: utf-8 -*-
"""Console scripts for threedigrid."""

from __future__ import absolute_import
import click
from IPython.terminal.embed import InteractiveShellEmbed

from threedigrid.admin.gridadmin import GridH5Admin
from threedigrid.management import help_texts

file_type_choice_map = {'shape': 'to_shape', 'gpkg': 'to_gpkg'}


@click.command()
@click.option('--grid-file', prompt='Path to the admin file',
              help='Path to the admin file', type=click.Path(exists=True))
@click.option('--ipy/--no-ipy', default=False,
              help='Start an interactive ipython session')
def kick_start(grid_file, ipy):
    grid = GridH5Admin(grid_file)

    txt = help_texts.model_overview(grid_file)
    click.secho("""\nOverview of model specifics:\n""", bold=True)
    click.secho(txt)

    if ipy:
        info_txt = '\n\nA GridH5Admin instance for model {} has been ' \
                   'created (variable named grid). Call dir(grid) to ' \
                   'explore.\n'.format(grid.model_name)
        click.secho(info_txt, bg='green', fg='black', bold=True)
        ipshell = InteractiveShellEmbed(exit_msg='Ciao...\n')
        ipshell()


@click.command()
@click.option('--grid-file', prompt='Path to the admin file',
              help='Path to the admin file', type=click.Path(exists=True))
@click.option('--file-type', type=click.Choice(['shape', 'gpkg']))
@click.option('--output-file', prompt='Path to the output file',
              help='Path to the output file',
              type=click.Path(writable=True, resolve_path=True))
@click.option('--model', type=click.Choice(
    ['nodes', 'lines', 'breaches', 'levees']))
@click.option('--subset', default='',
              help='Filter by a subset like 1D_all '
                   '(applies only for models nodes and lines')
def export_to(grid_file, file_type, output_file, model, subset):
    success_msg = 'Successfully created {} '.format(output_file)
    grid = GridH5Admin(grid_file)
    m = getattr(grid, model)
    export_func = file_type_choice_map[file_type]
    if subset:
        getattr(m.subset(subset), export_func)(output_file)
        click.secho(success_msg, fg='green', bold=True)
        return
    getattr(m, export_func)(output_file)
    click.secho(success_msg, fg='green', bold=True)


if __name__ == "__main__":
    kick_start()



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
"""
Models
++++++

The ``Nodes`` models represent the actual calculation points of the
threedicore and derived information like the cells they lie in.

For examples how to query the ``Nodes`` model see :ref:`api-label`


"""

from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
import numpy as np
from numba import jit

from threedigrid.orm.models import Model
from threedigrid.orm.fields import ArrayField
from threedigrid.orm.base.fields import BooleanArrayField
from threedigrid.orm.fields import PointArrayField
from threedigrid.orm.fields import BboxArrayField
from threedigrid.geo_utils import transform_xys
from threedigrid.numpy_utils import get_smallest_uint_dtype
from threedigrid.admin.nodes import exporters
from threedigrid.admin.nodes import subsets
from six.moves import zip


NODE_SUBSETS = {
    'node_type__eq': subsets.NODE_TYPE__EQ_SUBSETS,
    'node_type__in': subsets.NODE_TYPE__IN_SUBSETS,
}


class Nodes(Model):
    content_pk = ArrayField()
    seq_id = ArrayField()
    coordinates = PointArrayField()
    cell_coords = BboxArrayField()
    zoom_category = ArrayField()
    node_type = ArrayField()
    is_manhole = BooleanArrayField()
    sumax = ArrayField()

    SUBSETS = NODE_SUBSETS

    @property
    def connectionnodes(self):
        return self._filter_as(ConnectionNodes, content_pk__ne=0)

    @property
    def manholes(self):
        return self._filter_as(Manholes, is_manhole=True)

    @property
    def added_calculationnodes(self):
        return self._filter_as(AddedCalculationNodes, content_pk=0)

    def __init__(self, *args, **kwargs):

        super(Nodes, self).__init__(*args, **kwargs)

        self._exporters = [
            exporters.NodesOgrExporter(self),
        ]

    @property
    def locations_2d(self):
        data = self.subset('2D_open_water').data
        # x0 = 0, y0 = 0
        # Translate
        # data['coordinates'][0] += x0
        # data['coordinates'][1] += y0

        # Return [[node_id, coordinate_x + x0, coordinate_y + y0]]
        return list(zip(
            data['id'].tolist(),
            data['coordinates'][0].tolist(),
            data['coordinates'][1].tolist()))


class AddedCalculationNodes(Nodes):
    pass


class ConnectionNodes(Nodes):
    initial_waterlevel = ArrayField()
    storage_area = ArrayField()


class Manholes(ConnectionNodes):
    bottom_level = ArrayField()
    drain_level = ArrayField()
    display_name = ArrayField()
    surface_level = ArrayField()
    calculation_type = ArrayField()
    shape = ArrayField()
    width = ArrayField()
    manhole_indicator = ArrayField()


class Cells(Nodes):
    """
    Model that represents the threedicore calculation cells. ``Cells`` are
    a sub-class of ``Nodes`` because they share the same attributes.
    ``Cells``, however, also have a z_coordinate, the bottom level of
    the grid cell and cell_coords, the lower left and upper right coordinates
    of the cells extent.

    """

    z_coordinate = ArrayField()
    pixel_width = ArrayField()
    pixel_coords = BboxArrayField()

    def __init__(self, *args, **kwargs):

        super(Cells, self).__init__(*args, **kwargs)

        self._exporters = [
            exporters.CellsOgrExporter(self),
        ]

    @property
    def bounds(self):
        """
        :return: coordinates of the cell bounds (counter clockwise):
                 minx, miny, maxx, miny, maxx, maxy, minx, maxy
        """
        minx, miny, maxx, maxy = self.cell_coords
        return np.vstack(
            (minx, miny,
             maxx, miny,
             maxx, maxy,
             minx, maxy))

    def get_id_from_xy(self, x, y, xy_epsg_code=None, subset_name=None):
        """
        :param x: the x coordinate in xy_epsg_code
        :param y: the y coordinate in xy_epsg_code
        :param subset_name: filter on a subset of cells

        :return: numpy array with cell id's for x, y
        """

        if xy_epsg_code and xy_epsg_code != self.epsg_code:
            xy = transform_xys(
                np.array([x]), np.array([y]),
                xy_epsg_code, self.epsg_code).flatten()
        else:
            xy = [x, y]

        inst = self
        if subset_name:
            inst = self.subset(subset_name)
        id = inst.filter(cell_coords__contains_point=xy).id
        return id.tolist()

    def get_ids_from_pix_bbox(self, bbox, subset_name=None):
        """
        :param x: the x coordinate in xy_epsg_code
        :param y: the y coordinate in xy_epsg_code
        :param subset_name: filter on a subset of cells

        :return: numpy array with cell id's for x, y
        """

        inst = self
        if subset_name:
            inst = self.subset(subset_name)
        id = inst.filter(pixel_coords__in_bbox=bbox).id
        return id.tolist()

    def get_nodgrid(self, pix_bbox, subset_name=None):
        ids = np.array(
            self.get_ids_from_pix_bbox(pix_bbox, subset_name=subset_name),
            dtype=np.int32
        )
        return create_nodgrid(
            self.pixel_coords[:],
            ids[:],
            int(pix_bbox[2] - pix_bbox[0]),
            int(pix_bbox[3] - pix_bbox[1]),
            int(pix_bbox[0]),
            int(pix_bbox[1])
        )

    def __repr__(self):
        return "<orm cells instance of {}>".format(self.model_name)

    def __contenttype__(self):
        return 'cells'


class Grid(Model):
    """
    Implemented fields:

        - nodm
        - nodn
        - nodk

    They all have the same size as nodes and cells, which is why this
    model lives in the nodes/model module. In fact they are attributes
    of the cell coordinates
    """

    nodm = ArrayField()  #
    nodn = ArrayField()
    nodk = ArrayField()
    ip = ArrayField()
    jp = ArrayField()

    def __init__(self, *args, **kwargs):

        super(Grid, self).__init__(*args, **kwargs)
        self.n2dtot = kwargs['n2dtot']
        self.dx = kwargs['dx']

    def get_pixel_map(self, dem_pixelsize, dem_shape):
        """
        get the node grid to pixel map

        :param dem_pixelsize: pixelsize of the geo tiff
        :param dem_shape: shape of the numpy representation of the geo tiff

        :return: flipped array of the dem_shape that matches the geotiff
        """

        # Convert nod_grid to smallest uint type possible
        dtype = get_smallest_uint_dtype(maxval=self.n2dtot)
        grid_arr = np.zeros(dem_shape, dtype=dtype)

        # applies for for 2D nodes only
        _k = self.nodk[0:self.n2dtot + 1] - 1
        _m = self.nodm[0:self.n2dtot + 1] - 1
        _n = self.nodn[0:self.n2dtot + 1] - 1

        # the size in pixels of each grid cell
        _size = self.dx[_k] / dem_pixelsize
        # corresponding node index
        _ind = np.arange(0, self.n2dtot + 1, dtype='int')

        n_slice_start = np.array(_n * _size, dtype='int')
        n_slice_end = np.array((_n * _size) + _size, dtype='int')
        n_slice = list(zip(n_slice_start, n_slice_end))
        m_slice_start = np.array(_m * _size, dtype='int')
        m_slice_end = np.array((_m * _size) + _size, dtype='int')
        m_slice = list(zip(m_slice_start, m_slice_end))

        for ns, ms, idx in zip(n_slice, m_slice, _ind):
            axis_x = slice(*ns)
            axis_y = slice(*ms)
            grid_arr[axis_x, axis_y] = idx

        # flip upside-down to match geotiff
        return grid_arr[::-1, ::]


@jit(nopython=True)
def create_nodgrid(pixel_coords, ids, width, height, offset_i, offset_j):
    grid_arr = np.full((height, width), -9999, dtype=np.int32)
    for id in ids:
        i0 = np.maximum(pixel_coords[0, id], offset_i)
        i1 = np.minimum(pixel_coords[2, id], offset_i + width)
        j0 = np.maximum(pixel_coords[1, id], offset_j)
        j1 = np.minimum(pixel_coords[3, id], offset_j + height)
        grid_arr[j0: j1, i0: i1] = id
    return grid_arr



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
from collections import OrderedDict
from six.moves import range

LONLAT_DIGITS = 7  # 7 decimals is about 11mm accuracy

TYPE_V2_PIPE = 'v2_pipe'
TYPE_V2_CHANNEL = 'v2_channel'
TYPE_V2_CULVERT = 'v2_culvert'
TYPE_V2_ORIFICE = 'v2_orifice'
TYPE_V2_WEIR = 'v2_weir'
TYPE_V2_MANHOLE = 'v2_manhole'
TYPE_V2_PUMPSTATION = 'v2_pumpstation'
TYPE_V2_LEVEE = 'v2_levee'
TYPE_V2_1D_LATERAL = 'v2_1d_lateral'
TYPE_V2_WINDSHIELD = 'v2_windshielding'

TYPE_V2_1D_BOUNDARY_CONDITIONS = 'v2_1d_boundary_conditions'
TYPE_V2_2D_BOUNDARY_CONDITIONS = 'v2_2d_boundary_conditions'
TYPE_V2_CONNECTION_NODES = 'v2_connection_nodes'

TYPE_V2_BREACH = 'v2_breach'
TYPE_V2_CROSS_SECTION_DEF = 'v2_cross_section_definition'
TYPE_V2_CROSS_SECTION_LOCATION = 'v2_cross_section_location'
TYPE_V2_ADDED_CALCULATION_POINT = 'v2_added_calculation_point'


TYPE_CODE_MAP = {
    TYPE_V2_PIPE: 1,
    TYPE_V2_CHANNEL: 2,
    TYPE_V2_CULVERT: 3,
    TYPE_V2_ORIFICE: 4,
    TYPE_V2_WEIR: 5,
    TYPE_V2_MANHOLE: 6,
    TYPE_V2_PUMPSTATION: 7,
    TYPE_V2_LEVEE: 8,
    TYPE_V2_1D_LATERAL: 9,
    TYPE_V2_1D_BOUNDARY_CONDITIONS: 10,
    TYPE_V2_2D_BOUNDARY_CONDITIONS: 11,
    TYPE_V2_CONNECTION_NODES: 12,
    TYPE_V2_BREACH: 13,
    TYPE_V2_CROSS_SECTION_DEF: 14,
    TYPE_V2_CROSS_SECTION_LOCATION: 15,
    TYPE_V2_ADDED_CALCULATION_POINT: 16,
    TYPE_V2_WINDSHIELD: 17
}

# ------------------ HDF5 NAMES ------------------ #
DSET_ID_MAPPING = 'id_map'
GROUP_MAPPINGS = 'mappings'

DSET_GEOMS = 'geoms'
GROUP_LEVEE = 'levees'

###############################################################################
# HDF5 ATTRIBUTE NAMES
EXTENT_1D_KEY = 'extent_1d'
EXTENT_2D_KEY = 'extent_2d'
THREEDICORE_VERSION_KEY = 'threedicore_version'
HAS_1D_KEY = 'has_1d'
HAS_2D_KEY = 'has_2d'
HAS_PUMPSTATIONS_KEY = 'has_pumpstations'
HAS_BREACHES_KEY = 'has_breaches'


###############################################################################
# MANHOLE_TYPES is used by:
#
#   - v2_manhole

MANHOLE_TYPES = {
    0: 'manhole',
    1: 'outlet',
    2: 'pumpstation',
}


CALCULATION_TYPES = {
    # -1 tot 99 ::= Node | 'verbinding-zonder-subnodes(="vertices")'
    -1: 'boundary node',
    0: 'embedded',
    1: 'stand-alone',
    2: 'connected',
    3: 'broad crested',  # only orifices + weirs, corresponds
    4: 'short crested',  # only orifices + weirs
    5: 'double connected',

    # 100 tot oneindig ::= 'verbinding-met-subnodes(="vertices")'
    # a.k.a. pinpoint
    100: 'embedded',
    101: 'stand-alone',
    102: 'connected',
    105: 'double connected'
}

# Hardcoded zoom levels for pipe/nod/nod_1d/line/pump
# for 1-5 zoom level ("zoom_category") in sqlite file.


ZOOM_LEVELS = {
    'pipe': list(range(16, 21)),
    'nod': list(range(16, 21)),
    'nod_1d': list(range(16, 21)),
    'line': list(range(16, 21)),
    'pump': list(range(16, 21)),
}


###############################################################################
# SEWERAGE_TYPES is used by:
#
#   - v2_pipe

SEWERAGE_TYPES = {
    0: 'combined',
    1: 'stormwater',
    2: 'wastewater',
    3: 'transport',
    4: 'overflow',
    5: 'sinker',
    6: 'storage',
    7: 'storage settling tank',
    8: None,
}


###############################################################################
# CREST_TYPES is used by:
#
#   - v2_weir
#   - v2_orifice
#
# NB! This mapping is related to CALCULATION_TYPES mapping (key=3 | key=4)

CREST_TYPES = {
    3: 'broad crested',  # only orifices + weirs
    4: 'short crested',  # only orifices + weirs
}

# accounts for 2D only
LINE_BASE_FIELDS = OrderedDict([
    ('link_id', 'int'),
    ('kcu', 'int'),
    ('kcu_descr', 'str'),
    ('node_a', 'int'),
    ('node_b', 'int'),
])

# extends LINE_BASE_FIELDS in case of 1D
LINE_1D_FIELDS = OrderedDict([
    ('cont_type', 'str'),
    ('cont_pk', 'int'),
    ('seq_id', 'int'),
])

LINE_BASE_FIELDS_ALL = OrderedDict(LINE_BASE_FIELDS,  **LINE_1D_FIELDS)

# maps the fields names of grid line objects
# to their external representation
LINE_FIELD_NAME_MAP = OrderedDict([
    ('link_id', 'id'),
    ('kcu', 'kcu'),
    ('kcu_descr', 'kcu_descr'),
    ('node_a', 'node_a'),
    ('node_b', 'node_b'),
    ('seq_id', 'lik'),
    ('cont_type', 'content_type'),
    ('cont_pk', 'content_pk'),
])

# accounts for 2D only
NODE_BASE_FIELDS = OrderedDict([
    ('nod_id', 'int'),
])

NODE_1D_FIELDS = OrderedDict([
    ('con_nod', 'str'),
    ('con_nod_pk', 'int'),
])

NODE_FIELD_NAME_MAP = OrderedDict([
    ('nod_id', 'id'),
    ('con_nod', 'seq_id'),
    ('con_nod_pk', 'content_pk'),
])

TYPE_FUNC_MAP = {
    'int': int,
    'float': float,
    'str': str}


###############################################################################
# shapefile names (served by tilestache)
SUBSET_1D_ALL = '1D_all'
SUBSET_2D_GROUNDWATER = '2D_groundwater'
SUBSET_2D_OPEN_WATER = '2D_open_water'
SUBSET_2D_VERTICAL_INFILTRATION = '2D_vertical_infiltration'

# grids
_BASE_NAME = 'grid_'
GROUNDWATER = _BASE_NAME + SUBSET_2D_GROUNDWATER
OPEN_WATER = _BASE_NAME + SUBSET_2D_OPEN_WATER

# nodes
_NODES_BASE = 'nodes_'
NODES = _NODES_BASE + SUBSET_1D_ALL

# lines
_LINES_BASE = 'lines_'
LINES = _LINES_BASE + SUBSET_1D_ALL
GROUNDWATER_LINES = _LINES_BASE + SUBSET_2D_GROUNDWATER
OPEN_WATER_LINES = _LINES_BASE + SUBSET_2D_OPEN_WATER
VERTICAL_INFILTRATION_LINES = (
    _LINES_BASE + SUBSET_2D_VERTICAL_INFILTRATION)

# levees
LEVEES = "levees"


###############################################################################
# for extent calculations

# key: name of the node subset
# value: hdf5 attribute
SUBSET_NAME_H5_ATTR_MAP = {
    SUBSET_1D_ALL.upper(): EXTENT_1D_KEY,
    SUBSET_2D_OPEN_WATER.upper(): EXTENT_2D_KEY
}

NO_DATA_VALUE = -9999.

# the default slice for result timeseries
DEFAULT_CHUNK_TIMESERIES = slice(0, 10)



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import

import logging
from abc import ABCMeta
from collections import OrderedDict
from itertools import chain
from itertools import tee

import numpy as np
import six
from six.moves import zip

from threedigrid.orm.base.exceptions import OperationNotSupportedError
from threedigrid.orm.base.fields import ArrayField
from threedigrid.orm.base.fields import IndexArrayField
from threedigrid.orm.base.fields import TimeSeriesArrayField
from threedigrid.orm.base.filters import SliceFilter
from threedigrid.orm.base.filters import get_filter
from threedigrid.orm.base.options import Options

logger = logging.getLogger(__name__)


def extend_instance(obj, cls):
    """Apply mixins to a class instance after creation"""
    base_cls = obj.__class__
    base_cls_name = obj.__class__.__name__
    obj.__class__ = type(base_cls_name, (base_cls, cls), {})


def pairwise(iterable):
    # from https://docs.python.org/2/library/
    # itertools.html#recipes
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


class Model(six.with_metaclass(ABCMeta)):
    id = IndexArrayField()

    _field_names = []

    _exporters = []

    _filter_map = None

    _datasource = None

    def __init__(self, datasource=None, slice_filters=[],
                 epsg_code=None, only_fields=[], reproject_to_epsg=None,
                 has_1d=None, mixin=None, timeseries_chunk_size=None,
                 **kwargs):
        """
        Initialize a Model with a datasource, filters
        and a epsg_code.

        The datasource is a wrapped h5py Group from a H5py file or
        an instance that implements the same interface.
        """

        self._datasource = datasource
        self.slice_filters = slice_filters
        self.only_fields = only_fields
        self.reproject_to_epsg = reproject_to_epsg

        self.class_kwargs = {
            'slice_filters': slice_filters,
            'only_fields': only_fields,
            'reproject_to_epsg': reproject_to_epsg,
            'has_1d': has_1d,
            'mixin': mixin,
            'timeseries_chunk_size': timeseries_chunk_size
        }

        # Extend the class with the mixin, if set
        if mixin:
            extend_instance(self, mixin)
            # pass kwargs to mixin
            super(Model, self).__init__(**kwargs)

        # Cache the boolean filter mask for this instance
        # after it has been computed once
        self._boolean_mask_filter = None
        self._mixin = mixin
        self._epsg_code = epsg_code

        # Cache the field names
        _field_names = [
            x for x in dir(self.__class__)
            if isinstance(
                getattr(self.__class__, x),
                (ArrayField, TimeSeriesArrayField)
            )
        ]
        self._field_names = set(self._field_names).union(set(_field_names))

        self.has_1d = has_1d

    @property
    def epsg_code(self):
        if not self._epsg_code:
            self._epsg_code = self._datasource.getattr('epsg_code')
        return self._epsg_code

    @property
    def count(self):
        """count of all elements (including trash element)"""
        if self.class_kwargs.get('slice_filters'):
            return self.get_filtered_field_value('id').size
        return self.get_field_value('id').size

    @property
    def model_name(self):
        try:
            model_name = '-'.join(
                (self._datasource.getattr('model_name'),
                 self._datasource.getattr('model_slug'),
                 str(self._datasource.getattr('revision_nr')),
                 self._datasource.getattr('revision_hash')))
        except (AttributeError, KeyError, TypeError):
            model_name = 'unknown'
            pass
        return model_name

    def _get_field(self, field_name):
        """
        Returns: the ArrayField with field_name on this instance.
        """
        return super(Model, self).__getattribute__(field_name)

    def get_field_value(self, field_name, **kwargs):
        """
        Returns: the value for the ArrayField with name field_name
        on the instance.

        With this in place we can later override the ArrayField
        get_value method for field specific conversion, if needed.
        """
        update_dict = {
            'model_name': self.__class__.__name__,
        }

        kwargs.update(update_dict)
        return self._meta.get_field(field_name).get_value(
            self._datasource, field_name, **kwargs)

    def _get_subset_idx(self, field_name):
        """
        get an array of indexes for the given subset

        :param field_name: field name
        """
        new_inst = self.__init_class(
            self.__class__, **{'mixin': self.class_kwargs.get('mixin')})
        subset_dict = new_inst.Meta.subset_fields.get(field_name)
        if not subset_dict:
            return
        _subset_name = list(subset_dict.keys())
        if not _subset_name:
            return
        return new_inst.subset(_subset_name[0]).id

    def get_filtered_field_value(
            self, field_name, ts_filter=None, lookup_index=None,
            subset_index=None):
        # Redirect via datasource
        return self._datasource.get_filtered_field_value(
            self, field_name, ts_filter, lookup_index, subset_index)

    def __getattribute__(self, attr_name):
        """
        Override the __getattribute__ methode to return
        the computed values of ArrayField's instead of
        the ArrayField
        instance.
        """
        # Note: don't use getattr(self, XX) in this function,
        # it will lead to a max recursion error
        attr = super(Model, self).__getattribute__(attr_name)
        # Override getting ArrayField attributes
        if isinstance(attr, ArrayField):

            # Return meta directly, we don't want it to
            # to be filtered
            if attr_name == 'meta':
                return attr.get_value(self._datasource, attr_name)

            # Use the get function to retrieve the computed/filtered
            # value for the ArrayField with name: 'name'
            return super(Model, self).__getattribute__(
                'get_filtered_field_value')(attr_name)

        # Default behaviour, return the attribute from superclass
        return attr

    def __init_class(self, klass, **kwargs):
        """
        Returns: a new instance of 'klass' with new filters and
                 same datasource.
        """

        return klass(
            datasource=self._datasource,
            **kwargs)

    def __get_filters(self, **kwargs):
        """
        Parse kwargs for filters.

        For example:
            content_type__in=[1,2,3]

            splitted_key[0] = "content_type"
            splitted_key[1] = "in"
            -> Try to find filter with "in" as key

        """
        new_slice_filters = list(self.slice_filters)  # make copy
        filter_as = kwargs.pop('filter___as', False)

        for key, value in six.iteritems(kwargs):
            # python2/3 combat
            if isinstance(value, str):
                value = str.encode(value)
            splitted_key = key.split('__')
            if splitted_key[0] not in self._field_names:
                raise ValueError(
                    "Field '{}' unknown. Choices are {}".format(
                        splitted_key[0],
                        self._field_names)
                )

            new_slice_filters.append(
                get_filter(
                    splitted_key,
                    self._meta.get_field(splitted_key[0]),
                    value,
                    filter_map=self._filter_map,
                    filter_as=filter_as)
            )

        return new_slice_filters

    def _filter_as(self, klass, **kwargs):
        """
        Same as self.filter, but now return "klass" instance instead of
        self.__class__ instance
        """
        # Inject filter___as, set this on the filters
        # so they can identified as filters that
        # produce submodels
        kwargs['filter___as'] = True
        slice_filters = self.__get_filters(**kwargs)
        new_class_kwargs = dict(self.class_kwargs)
        new_class_kwargs.update(
            {'slice_filters': slice_filters})

        return self.__init_class(klass, **new_class_kwargs)

    # def _set_subsets(self):
    #     if not hasattr(self, 'Meta'):
    #         return
    #     if not hasattr(self.Meta, 'subset_fields'):
    #         return
    #
    #     _subsets = set(self.Meta.subset_fields.keys())
    #     for sset in _subsets:
    #         idx = self._get_subset_idx(sset)
    #         setattr(self, sset, idx)

    def filter(self, **kwargs):
        """
        Django style filtering, allows multiple filter or chaining filters
        for example:

            line_instance.filter(content_type='v2_weir', content_pk=1)
            line_instance.filter(content_type='v2_weir').filter(content_pk=1)

            Both produce the same result object.


        Returns: a new instance of the object with the extra filters added.
        """
        slice_filters = self.__get_filters(**kwargs)
        new_class_kwargs = dict(self.class_kwargs)
        new_class_kwargs.update(
            {'slice_filters': slice_filters})

        return self.__init_class(
            self.__class__, **new_class_kwargs)

    def slice(self, start, stop=None, step=None, override_filter_error=False):
        """
        slice by name or slice. See instance.predefined_slices for an overview
        of predefined slices

        :param s: either the name of the slice (see instance.predefined_slices
             for an overview or an slice instance, e.g. slice(<x>, <y>)
        :return: nodes instance with containing the sliced dataset.Example
            usage::
            usage:
            open_water_2d = nodes.slice('2D_open_water')
            open_water_2d.data

                OrderedDict([('content_pk', array([1, 2, 3, ...

        """
        if self.slice_filters and not override_filter_error:
            raise OperationNotSupportedError(
                'You cannot use slices on a filtered dataset')

        if isinstance(start, slice):
            slice_filter = start
        else:
            slice_filter = slice(start, stop, step)

        new_class_kwargs = dict(self.class_kwargs)
        new_class_kwargs.update(
            {'slice_filters':
             [SliceFilter(slice_filter)] + self.slice_filters})

        return self.__init_class(
            self.__class__, **new_class_kwargs)

    @property
    def known_subset(self):
        if not hasattr(self, 'SUBSETS'):
            return "has no subsets defined"
        return list(
            chain(*[list(v.keys()) for v in six.itervalues(self.SUBSETS)])
        )

    def subset(self, name):
        """
        get a subset of the data. See instance.subsets for an overview
        of predefined subsets

        :param name: name of the subset (see instance.predefined_slices
             for an overview or an slice instance, e.g. slice(<x>, <y>)
        :return: nodes instance with containing the sliced dataset.Example
            usage::
            open_water_2d = nodes.slice('2D_open_water')
            open_water_2d.data

                OrderedDict([('content_pk', array([1, 2, 3, ...

        """
        if not hasattr(self, 'SUBSETS'):
            raise TypeError("SUBSETS not defined for this type of model")

        if isinstance(name, six.string_types):
            field_filter = [key for key in self.SUBSETS if
                            name.upper() in self.SUBSETS[key]]

            if not field_filter:
                msg = 'Invalid subset name.'
                msg += ' Valid options are: %s ' % (
                    [x for y in self.SUBSETS.values() for x in y], )
                logger.exception(msg)
                raise KeyError(msg)

            if len(field_filter) > 1:
                raise KeyError(
                    'Multiple options found for subset: %s ' % name.upper())

            field_filter = field_filter[0]
        else:
            raise TypeError('Type %s not supported' % type(name,))

        return self.filter(
            **{field_filter: self.SUBSETS[field_filter][name.upper()]})

    def only(self, *args):
        """
        Return only the fields requested

        Usage: line_instance.only('content_pk', 'line')
        """
        if not args:
            raise Exception("Please provide at least one field name")

        for x in args:
            if x not in self._field_names:
                raise Exception("Unknown field name: {0}".format(x))

        new_class_kwargs = dict(self.class_kwargs)
        new_class_kwargs.update(
            {'only_fields': self.only_fields + list(args)})

        return self.__init_class(
            self.__class__, **new_class_kwargs)

    def _get_values(self, selection):
        """
        when no filters are specified, we are still operating on the hdf5
        groups so get the values explicitly

        :param selection: ordered dict of selected fields
            with group datasets
        :return: new ordered dict with values from group datasets
        """

        _tmp = OrderedDict()
        for k, v in six.iteritems(selection):
            try:
                _tmp[k] = v[:]
            except TypeError:
                pass
        return _tmp

    def to_dict(self):
        """
        Returns: the filtered values as dictionairy
        """
        return self._datasource.execute_query(self)

    def to_structured_array(self):
        """
        :return: the filtered values as a
        structured (named) array
        """
        selection = self.to_dict()

        # Convert the dictionary to structured array
        dtypes = []
        for key, value in selection.items():
            dtypes.append((key, value.dtype.name, value.shape))

        return np.array(
            [tuple([selection[x[0]] for x in dtypes])],
            dtype=dtypes)[0]

    def to_list(self):
        """
        :return: list of dicts with key's and values
        """

        # Filter results and transform the result to
        # and np.ndarray
        selection = self.to_dict()
        if len(list(selection.values())) > 1:
            array = np.array(list(selection.values()))
        else:
            array = list(selection.values())[0]

        def optional_zip(array_to_zip):
            """
            Optional zip the results of
            array_to_zip if there are more than one
            dimensions.
            """
            # Transform data to list
            array_as_list = array_to_zip.tolist()
            if len(array_to_zip.shape) > 1:
                # Zip the list if there are more than
                # one dimensions.
                array_as_list = zip(*array_as_list)
            return array_as_list

        # Zip the lists of array_lists (does same as a transpose)
        data = zip(*[optional_zip(x) for x in array])

        # Create a list of dictionairies of the data by
        # zipping the selection.keys() (= field_names) for all items
        # in data
        return [dict(zip(list(selection.keys()), x)) for x in data]

    @property
    def boolean_mask_filter(self):
        selection = OrderedDict()

        # Compute the boolean mask filter
        # that can be applied to all array's
        # based on the specified filters

        if self._boolean_mask_filter is None:

            # If no filters are defined
            if not self.slice_filters:
                # Select everything
                return slice(None)

            # Only load the fields that need to be filtered
            filter_field_names = [
                x.get_field_name() for x in self.slice_filters]

            for name in [x for x in filter_field_names if x]:
                selection[name] = self.get_field_value(
                    name)

            boolean_mask_filter = None
            # Compute the the filter
            for filter_instance in self.slice_filters:
                filter_bool_mask = filter_instance.get_boolean_mask_filter(
                        selection, self)

                if boolean_mask_filter is None:
                    boolean_mask_filter = filter_bool_mask
                else:
                    # apply the filter_bool_mask on the boolean_mask_filter
                    # where the boolean_mask_filter is still "True".
                    boolean_mask_filter[
                        boolean_mask_filter == True] &= filter_bool_mask # noqa

            self._boolean_mask_filter = boolean_mask_filter

        return self._boolean_mask_filter

    def to_array(self):
        """
        Returns: the filtered values as a numpy array
        """
        selection = self.to_dict()
        if len(list(selection.values())) > 1:
            return np.array(list(selection.values()))
        return list(selection.values())[0]

    @property
    def data(self):
        """
        Returns: the (filtered) data a dictionairy
        """
        return self.to_dict()

    @property
    def _meta(self):
        """
        meta class that add meta data to a ResultMixin instance. The meta data
        includes entries for the ``_meta_fields`` collection (if found).

            >>> from threedigrid.admin.gridresultadmin import GridH5ResultAdmin
            >>> f = "/code/tests/test_files/results_3di.nc"
            >>> ff = "/code/tests/test_files/gridadmin.h5"
            >>> gr = GridH5ResultAdmin(ff, f)
            >>> gr.nodes._meta.s1
            >>> s1(units=u'm', long_name=u'waterlevel', standard_name=u'water_surface_height_above_reference_datum') # noqa
        """
        return Options(self)

    def __repr__(self):
        """human readable representation of the instance"""
        return "<orm {} instance of {}>".format(
            self.__contenttype__(), self.model_name
        )

    def __contenttype__(self):
        """conent type, e.g. lines, nodes, cells, ..."""
        return self._datasource.group_name



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
import os
import logging
from collections import OrderedDict
import six
from six.moves import range

try:
    from osgeo import ogr
except ImportError:
    ogr = None

from threedigrid.orm.base.exporters import BaseOgrExporter

from threedigrid.geo_utils import get_spatial_reference
from threedigrid.admin import exporter_constants as const
from threedigrid.admin.constants import NODE_BASE_FIELDS
from threedigrid.admin.constants import NODE_1D_FIELDS
from threedigrid.admin.constants import NODE_FIELD_NAME_MAP
from threedigrid.admin.constants import TYPE_FUNC_MAP

logger = logging.getLogger(__name__)


class NodesOgrExporter(BaseOgrExporter):
    """
    Exports to ogr formats. You need to set the driver explicitly
    before calling save()
    """
    def __init__(self, nodes):
        """
        :param lines: lines.models.Lines instance
        """
        self._nodes = nodes
        self.supported_drivers = {
            const.GEO_PACKAGE_DRIVER_NAME,
            const.SHP_DRIVER_NAME,
            const.GEOJSON_DRIVER_NAME,
        }

    def save(self, file_name, node_data, target_epsg_code, **kwargs):
        """
        save to file format specified by the driver, e.g. shapefile

        :param file_name: name of the outputfile
        :param node_data: dict of node data
        """
        assert self.driver is not None
        geomtype = 0
        sr = get_spatial_reference(target_epsg_code)
        self.del_datasource(file_name)
        data_source = self.driver.CreateDataSource(file_name)
        layer = data_source.CreateLayer(
            str(os.path.basename(file_name)),
            sr,
            geomtype
        )
        fields = NODE_BASE_FIELDS
        if self._nodes.has_1d:
            fields.update(NODE_1D_FIELDS)

        for field_name, field_type in six.iteritems(fields):
            layer.CreateField(ogr.FieldDefn(
                    str(field_name),
                    const.OGR_FIELD_TYPE_MAP[field_type])
            )
        _definition = layer.GetLayerDefn()

        for i in range(node_data['id'].size):
            point = ogr.Geometry(ogr.wkbPoint)
            point.AddPoint(
                node_data['coordinates'][0][i],
                node_data['coordinates'][1][i]
            )
            feature = ogr.Feature(_definition)
            feature.SetGeometry(point)
            for field_name, field_type in six.iteritems(fields):
                fname = NODE_FIELD_NAME_MAP[field_name]
                raw_value = node_data[fname][i]
                value = TYPE_FUNC_MAP[field_type](raw_value)
                feature.SetField(str(field_name), value)
            layer.CreateFeature(feature)


class CellsOgrExporter(BaseOgrExporter):
    """
    Exports to ogr formats. You need to set the driver explicitly
    before calling save()
    """
    def __init__(self, cells):
        """
        :param lines: lines.models.Lines instance
        """
        self._cells = cells
        self.supported_drivers = {
            const.GEO_PACKAGE_DRIVER_NAME,
            const.SHP_DRIVER_NAME,
        }
        self.driver = None

    def save(self, file_name, cells_data, target_epsg_code, **kwargs):
        """
        save to file format specified by the driver, e.g. shapefile

        :param file_name: name of the outputfile
        :param cells_data: dict of cell data
        """
        assert self.driver is not None

        geomtype = 0
        sr = get_spatial_reference(target_epsg_code)

        self.del_datasource(file_name)
        data_source = self.driver.CreateDataSource(file_name)
        layer = data_source.CreateLayer(
            str(os.path.basename(file_name)),
            sr,
            geomtype
        )
        fields = OrderedDict([
            ('nod_id', 'int'),
            ('bottom_lev', 'float'),
        ])
        for field_name, field_type in six.iteritems(fields):
            layer.CreateField(
                ogr.FieldDefn(
                    str(field_name),
                    const.OGR_FIELD_TYPE_MAP[field_type]
                )
            )

        _definition = layer.GetLayerDefn()
        for i in range(cells_data['id'].size):
            feature = ogr.Feature(_definition)
            # Create ring
            ring = ogr.Geometry(ogr.wkbLinearRing)
            ring.AddPoint(cells_data['cell_coords'][0][i],
                          cells_data['cell_coords'][1][i])

            ring.AddPoint(cells_data['cell_coords'][2][i],
                          cells_data['cell_coords'][1][i])

            ring.AddPoint(cells_data['cell_coords'][2][i],
                          cells_data['cell_coords'][3][i])

            ring.AddPoint(cells_data['cell_coords'][0][i],
                          cells_data['cell_coords'][3][i])

            ring.AddPoint(cells_data['cell_coords'][0][i],
                          cells_data['cell_coords'][1][i])

            # Create polygon from ring
            poly = ogr.Geometry(ogr.wkbPolygon)
            poly.AddGeometry(ring)
            feature.SetGeometry(poly)
            feature.SetField(str("nod_id"), int(cells_data['id'][i]))
            feature.SetField(
                str("bottom_lev"), float(cells_data['z_coordinate'][i])
            )
            layer.CreateFeature(feature)



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
import numpy as np
# optional install results
try:
    from cftime import num2date
except ImportError:
    pass

from threedigrid.orm.base.fields import TimeSeriesCompositeArrayField
from threedigrid.orm.base.fields import TimeSeriesSubsetArrayField
from threedigrid.orm.base.fields import TimeSeriesArrayField
from threedigrid.orm.base.utils import _flatten_dict_values

import six


class ResultMixin(object):
    """
    Subclass this mixin and add the result
    fields as 'TimeSeriesArrayField'
    """
    timeseries_mask = None
    _done_composition = False

    def __init__(self, *args, **kwargs):
        # pop mixin specific fields and store them
        # in the class_kwargs
        self.timeseries_filter = kwargs.pop('timeseries_filter', None)
        self.timeseries_sample = kwargs.pop("timeseries_sample", None)
        self.timeseries_mask = None
        self.class_kwargs.update({
            'timeseries_filter': self.timeseries_filter,
            'timeseries_sample': self.timeseries_sample})
        if not self._done_composition and hasattr(self, 'Meta'):
            self._set_composite_fields()
            self._set_subset_fields()

    def _set_composite_fields(self):
        """
        add composite fields to the instance
        """

        if not hasattr(self.Meta, 'composite_fields'):
            return

        fields = {
            v: TimeSeriesCompositeArrayField(meta=self.Meta)
            for v in self.Meta.composite_fields.keys()
        }
        self._meta.add_fields(fields, hide_private=True)
        self._done_composition = True

    def _set_subset_fields(self):
        """
        add subset fields to the instance
        """
        if not hasattr(self, 'Meta'):
            return

        if not hasattr(self.Meta, 'subset_fields'):
            return

        fields = {}
        id_value = self.get_field_value('id')
        if id_value:
            count = self.get_field_value('id').size
        else:
            # Dummy value
            count = 1024
        for v, k in six.iteritems(self.Meta.subset_fields):
            _source_name = _flatten_dict_values(k)
            if not _source_name:
                continue
            source_name = _source_name[0]

            fields[v] = TimeSeriesSubsetArrayField(
                source_name=source_name, size=count
            )

        self._meta.add_fields(fields, hide_private=True)

    def generate_timeseries_mask(
            self, start_time=None, end_time=None, indexes=None):
        timestamps = self.timestamps
        timeseries_mask = True

        if all((start_time is None, end_time is None, indexes is None)):
            raise KeyError(
                "Please provide either start_time, end_time or indexes")

        if any((start_time is not None, end_time is not None)):
            if start_time is not None:
                timeseries_mask &= timestamps >= start_time
            if end_time is not None:
                timeseries_mask &= timestamps <= end_time

            self.timeseries_mask = timeseries_mask
        else:
            if isinstance(indexes, list) or isinstance(indexes, tuple):
                self.timeseries_mask = np.array(indexes)
            elif isinstance(indexes, slice):
                self.timeseries_mask = indexes
            else:
                raise TypeError(
                    "indexes should either be a list/tuple or a slice")

        if self.timeseries_sample:
            num_points = self.timeseries_sample['num_points']
            # Only sample if the required num_points is smaller
            # than the available points
            timestamps_size = self.timestamps.size
            if timestamps_size > num_points:
                if isinstance(self.timeseries_mask, slice):
                    indexes = np.arange(
                        timestamps_size
                    )[self.timeseries_mask].flatten().tolist()
                else:
                    indexes = np.argwhere(
                        self.timeseries_mask).flatten().tolist()

                if hasattr(self._datasource, 'swmr_mode'):
                    # always exclude the last item in swmr_mode, to be sure
                    # that the datasource['time'].size is the same
                    # as of the datasource['xxx'].size datasets that
                    # are going to be sampled.
                    indexes = indexes[:-1]

                self.timeseries_mask = self._get_indexes_subset(
                    indexes,
                    limit=self.timeseries_sample['num_points'],
                    include_end=self.timeseries_sample['include_end'])

    def _get_indexes_subset(self, indexes, limit, include_end=True):
        """
        Get indexes subset with length limit
        """

        # only limit if needed
        if len(indexes) <= limit:
            return indexes

        start = indexes[0]
        end = indexes[-1]
        divider = (end - start) / float(limit)
        indexes = []

        if include_end:
            divider = -divider
            x = end
        else:
            x = start

        while len(indexes) < limit:
            indexes.append(round(x))
            x += divider

        if include_end:
            indexes = indexes[::-1]

        return indexes

    def sample(self, num_points=None, include_end=True):
        """
        Sample the requested timeseries and return at
        maximum 'num_points' amount of points.

        :param num_points: the amount of sample points to return
        :param include_end: if true, the last point is always included,
                            else the first point is always included.
        """
        new_class_kwargs = dict(self.class_kwargs)
        new_class_kwargs.update({
            'timeseries_sample': {
                'num_points': num_points,
                'include_end': include_end}
        })

        return self.__class__(
            datasource=self._datasource,
            **new_class_kwargs)

    def timeseries(
            self, start_time=None, end_time=None, indexes=None):
        """
        Allows filtering on timeseries.

        You can either filter by start_time and end_time or indexes.

        Example usage for start_time and end_time filter::

            >>> from threedigrid.admin.gridresultadmin import GridH5ResultAdmin
            >>> nc = "/code/tests/test_files/results_3di.nc"
            >>> f = "/code/tests/test_files/gridadmin.h5"
            >>> gr = GridH5ResultAdmin(f, nc)
            >>> gr.nodes.timeseries(start_time=0, end_time=10).s1

        Example usage for index filter::

            >>> from threedigrid.admin.gridresultadmin import GridH5ResultAdmin
            >>> nc = "/code/tests/test_files/results_3di.nc"
            >>> f = "/code/tests/test_files/gridadmin.h5"
            >>> gr = GridH5ResultAdmin(f, nc)
            >>> qs_s1 = gr.nodes.timeseries(indexes=[1, 2, 3]).s1
            >>> qs_s1.shape
            >>> (3, 25156)

        A more complex example combining filters::

            >>> breaches_u1 = gr.lines.filter(kcu__eq=55).timeseries(indexes=[10, 11, 12, 13]).u1  # noqa
            >>> breaches_u1.shape
            >>> (4, 16)

        Filtering by indexes works with both lists (like in the examples above)
        and slices. This code yields exactly the same result

            >>> breaches_u1_slice = gr.lines.filter(kcu__eq=55).timeseries(indexes=slice(10,14)).u1  # noqa
            >>> breaches_u1_slice.shape
            >>> (4, 16)
            >>> assert np.all(breaches_u1 == breaches_u1_slice)

        :return: new instance with filtering options enabled
        """

        # self.generate_timeseries_mask(start_time, end_time, indexes)

        # Create a copy of the class_kwargs
        # and update it with timeseries_mask
        new_class_kwargs = dict(self.class_kwargs)
        new_class_kwargs.update({
            'timeseries_filter': {
                'start_time': start_time,
                'end_time': end_time,
                'indexes': indexes}
        })

        return self.__class__(
            datasource=self._datasource,
            **new_class_kwargs)

    def get_timeseries_mask_filter(self):
        """
        :return: the timeseries mask to be used for filtering
                 on timeseries
        """
        if self.timeseries_mask is None and self.timeseries_filter is not None:
            self.generate_timeseries_mask(
                **self.timeseries_filter
            )
        if self.timeseries_mask is not None:
            return self.timeseries_mask
        return self.class_kwargs.get('timeseries_chunk_size')

    @property
    def timestamps(self):
        """
        Get the list of timestamps for the results
        """

        # override if datasource has get_timestamps function
        if hasattr(self._datasource, 'get_timestamps'):
            return self._datasource.get_timestamps()

        time_key = 'time'
        if time_key not in list(self._datasource.keys()):
            raise AttributeError(
                'Result {} has no attribute {}'.format(
                    self._datasource.netcdf_file.filepath(), time_key)
            )

        value = self._datasource['time'][:]
        if self.timeseries_mask is not None:
            value = value[self.timeseries_mask]
        return value

    def get_timestamps(self, timeseries_mask):
        time_key = 'time'
        if time_key not in list(self._datasource.keys()):
            raise AttributeError(
                'Result {} has no attribute {}'.format(
                    self._datasource.netcdf_file.filepath(), time_key)
            )
        value = self._datasource['time'][:]
        return value[timeseries_mask]

    @property
    def dt_timestamps(self):
        return [
            t.isoformat() for t in num2date(
                self.timestamps,
                units=self._datasource['time'].attrs.get(
                    'units').decode('utf-8'))
        ]


class AggregateResultMixin(ResultMixin):
    """
    Subclass this mixin and add the result
    fields as 'TimeSeriesArrayField' or TimeSeriesCompositeArrayField
    """
    def __init__(self, *args, **kwargs):
        super(AggregateResultMixin, self).__init__(*args, **kwargs)

        # Create a copy of the class_kwargs
        # and update it with timeseries_mask
        new_class_kwargs = dict(self.class_kwargs)
        new_class_kwargs.update({
            'timeseries_mask': self.timeseries_mask})

    def get_timeseries_mask_filter(self):
        """
        :return: the timeseries mask to be used for filtering
                 on timeseries
        """
        if self.timeseries_filter is not None:
            start_time = self.timeseries_filter['start_time']
            end_time = self.timeseries_filter['end_time']
            indexes = self.timeseries_filter['indexes']

            self.timeseries_mask = {}
            field_names = self._meta.get_fields()
            for field_name, field_inst in six.iteritems(field_names):
                if not isinstance(field_inst, TimeSeriesArrayField):
                    continue
                ts = self.get_timestamps(field_name)
                mask = self._get_mask(start_time, end_time, indexes, ts)
                self.timeseries_mask[field_name] = mask
            return self.timeseries_mask
        return None

    def timeseries(self, start_time=None, end_time=None, indexes=None):
        """
        Allows filtering on timeseries.

        :param start_time: start_time in seconds
        :param end_time: end_time in seconds
        :param indexes: a slice, e.g. slice(<start>, <stop>, <step>)

        You can either filter by start_time and end_time or indexes. Indexes,
        unlike the GridH5ResultAdmin timeseries filter, allows only for slices
        when used with aggregated results.

        Example usage for start_time and end_time filter::

            >>> from threedigrid.admin.gridresultadmin import GridH5AggregateResultAdmin # noqa
            >>> nc = "/code/tests/test_files/aggregate_results_3di.nc"
            >>> f = "/code/tests/test_files/gridadmin.h5"
            >>> gr = GridH5AggregateResultAdmin(f, nc)
            >>> gr.nodes.timeseries(start_time=0, end_time=800).s1_max

        Example usage for index filter::

            >>> from threedigrid.admin.gridresultadmin import GridH5AggregateResultAdmin # noqa
            >>> nc = "/code/tests/test_files/aggregate_results_3di.nc"
            >>> f = "/code/tests/test_files/gridadmin.h5"
            >>> gr = GridH5AggregateResultAdmin(f, nc)
            >>> qs_s1 = gr.nodes.timeseries(indexes=slice(0,3)).s1
            >>> qs_s1.shape[0]
            >>> 3

        """
        new_class_kwargs = dict(self.class_kwargs)
        new_class_kwargs.update({
            'timeseries_filter': {
                'start_time': start_time,
                'end_time': end_time,
                'indexes': indexes}
        })
        return self.__class__(
            datasource=self._datasource,
            **new_class_kwargs)

    def _get_mask(self, start_time, end_time, indexes, timestamps):
        """

        :param start_time: start_time in seconds
        :param end_time: end_time in seconds
        :param indexes: a slice, e.g. slice(<start>, <stop>, <step>)
        :param timestamps: array of timestamps in seconds
        :return:
        """
        if all((start_time is None, end_time is None, indexes is None)):
            raise KeyError(
                "Please provide either start_time, end_time or indexes")

        timeseries_mask = True
        if any((start_time is not None, end_time is not None)):
            if start_time is not None:
                timeseries_mask &= timestamps >= start_time
            if end_time is not None:
                timeseries_mask &= timestamps <= end_time

            timeseries_mask = timeseries_mask
        else:
            if isinstance(indexes, slice):
                timeseries_mask = indexes
            else:
                raise TypeError(
                    "indexes should either be a slice")

        return timeseries_mask

    @property
    def timestamps(self):
        timestamps_dict = {}

        for field_name, field in self._meta.get_fields().items():
            if isinstance(field, TimeSeriesArrayField):
                try:
                    timestamps_dict[field_name] = self.get_timestamps(
                        field_name)
                except AttributeError:
                    timestamps_dict[field_name] = np.array([])

        return timestamps_dict

    def get_timestamps(self, field_name):
        """
        Get the array of timestamps for a result field

        :param field_name: name of the field
        :return: array of timestamps
        :raises AttributeError when no the field has no timestamps
        """

        time_key = 'time_' + field_name

        if time_key in list(self._datasource.keys()):
            field_timestamps = self._datasource[time_key][:]
            # Mask the field_timestamps with the timeseries_mask of
            # that field when available
            if ((hasattr(self, 'timeseries_mask') and
                 self.timeseries_mask is not None)):
                if field_name in self.timeseries_mask:
                    field_timestamps = field_timestamps[
                        self.timeseries_mask[field_name]]
            return field_timestamps
        raise AttributeError('No timestamps found for {}'.format(field_name))

    def get_time_unit(self, field_name):
        """
        Get the time unit for a result field
        :param field_name: name of the field
        :return: <string> time unit
        :raises AttributeError when the field has no time unit attribute
        """

        time_key = 'time_' + field_name

        if time_key in list(self._datasource.keys()):
            arr = self._datasource[time_key]
            try:
                return arr.attrs.get('units')
            except AttributeError:
                raise AttributeError(
                    'No time unit found for field {}'.format(field_name)
                )



# -*- coding: utf-8 -*-

"""Unit test package for threedigrid."""
from __future__ import absolute_import
import threedigrid
import six


def test_version():
    v = threedigrid.__version__
    assert isinstance(v, six.string_types) and len(v) > 0



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
import logging
import numpy as np
from threedigrid.orm.base.datasource import DataSource
from collections import OrderedDict
from h5py._hl.dataset import Dataset
from threedigrid.admin.h5py_swmr import H5SwmrFile


logger = logging.getLogger(__name__)


class H5pyGroup(DataSource):
    """
    Datasource wrapper for h5py groups,
    adding meta data to all groups.
    """
    _group_name = None
    _h5py_file = None

    def get_filtered_field_value(
            self, model, field_name, ts_filter=None, lookup_index=None,
            subset_index=None):

        kwargs = {}
        if ts_filter is None:
            if hasattr(model, 'get_timeseries_mask_filter'):
                timeseries_filter = model.get_timeseries_mask_filter()
                ts_filter = timeseries_filter
                if isinstance(timeseries_filter, dict):
                    ts_filter = timeseries_filter.get(field_name)

        if ts_filter is not None:
            kwargs.update(
                {'timeseries_filter': ts_filter}
            )

        if lookup_index is None:
            if model._mixin and hasattr(model.Meta, 'lookup_fields'):
                lookup_index = model._meta._get_lookup_index()

        if lookup_index is not None:
            kwargs.update({'lookup_index': lookup_index})

        if subset_index is None:
            if model._mixin and hasattr(model.Meta, 'subset_fields'):
                subset_index = model._get_subset_idx(field_name)

        if subset_index is not None:
            kwargs.update({'subset_index': model._get_subset_idx(field_name)})

        value = model.get_field_value(field_name, **kwargs)

        # Transform the base_filter by prepending slice(None) to
        # match the dimensionality of the nparray_dict[key].shape
        #
        #      shape(100,) => _filter = [base_filter]
        #      shape(2, 100)  => _filter = [slice(None), base_filter]
        #
        #      Note: x[slice(None),[1,2,3]] == x[:,[1,2,3]]
        # if hasattr(self, 'get_timeseries_mask_filter'):
        #     timeseries_filter = self.get_timeseries_mask_filter()
        # else:
        #     timeseries_filter = slice(None)

        # Return a numpy array with None as only element when
        # the value is None.
        if value is None or value.size == 0:
            return np.array([])

        _filter = [slice(None)] * (
            len(value.shape) - 1) + [model.boolean_mask_filter]

        # By default load all data from H5,
        # this is WAY much faster
        if isinstance(value, Dataset):
            value = value[:]

        # Perform slicing by applying the mask
        value = value[tuple(_filter)]

        # Reproject any coordinates if a reproject_to_epsg is set and
        # there are coordinatefields in the selection
        if model.reproject_to_epsg and model._is_coords(field_name):
            value = model._do_reproject_value(
                    value, field_name, model.reproject_to_epsg)

        if isinstance(value, np.ma.core.MaskedArray):
            # Always return the data of a masked array
            value = value.data

        return value

    def execute_query(self, model):
        """
        Process a query on the model
        """
        selection = OrderedDict()

        ts_filter = None
        timeseries_filter = None
        lookup_index = None
        has_subsets = False

        if hasattr(model, 'get_timeseries_mask_filter'):
            timeseries_filter = model.get_timeseries_mask_filter()

        if model._mixin and hasattr(model.Meta, 'lookup_fields'):
            lookup_index = model._meta._get_lookup_index()

        if model._mixin and hasattr(model.Meta, 'subset_fields'):
            has_subsets = True

        for n in model._field_names:
            if not model.only_fields or n in model.only_fields:
                if isinstance(timeseries_filter, dict):
                    ts_filter = timeseries_filter.get(n)
                else:
                    ts_filter = timeseries_filter

                if has_subsets:
                    subset_index = model._get_subset_idx(n)
                else:
                    subset_index = None

                selection[n] = model.get_filtered_field_value(
                    n, ts_filter=ts_filter, lookup_index=lookup_index,
                    subset_index=subset_index)

        # Inject timestamps automatically
        if hasattr(model, 'get_timestamps'):
            selection['timestamps'] = model.get_timestamps(timeseries_filter)

        return selection

    def __init__(self, h5py_file, group_name, meta=None, required=False):
        self.group_name = group_name
        self._h5py_file = h5py_file

        if group_name not in list(h5py_file.keys()) and not required:
            logger.info(
                '[*] {0} not found in file {1}, not required...'.format(
                    group_name, h5py_file)
            )
            return

        try:
            self._source = h5py_file.require_group(group_name)
        except TypeError:
            self._source = h5py_file[group_name]

        self.meta = dict(
            [(y, [x[()]]) for y, x in self._h5py_file.get('meta').items()])
        self.meta['trash'] = [1]

    def set(self, name, values):
        if name in self._source:
            # Overwrite
            self._source[name][:] = values
        else:
            # Create
            self._source.create_dataset(name, data=values)

    def getattr(self, name):
        attr = self._h5py_file.attrs[name]
        if isinstance(attr, bytes):
            attr = attr.decode('utf-8')
        return attr

    def keys(self):
        if self._source is None:
            return []
        return list(self._source.keys())

    def has_any(self, sources):
        """
        Check if the any of the sources can be found in
        the file
        """
        return any(x in list(self.keys())
                   for x in sources)


class H5pyResultGroup(H5pyGroup):
    def __init__(self, h5py_file, group_name, netcdf_file,
                 meta=None, required=False):
        super(H5pyResultGroup, self).__init__(
            h5py_file, group_name, meta, required)
        self.netcdf_file = netcdf_file

        if isinstance(netcdf_file, H5SwmrFile):
            self.swmr_mode = True

    def keys(self):
        keys = list(super(H5pyResultGroup, self).keys())
        keys += list(self.netcdf_file.keys())
        return list(set(keys))

    def get(self, name):
        # meta is special
        if name == 'meta':
            return self.meta

        if name in self._source:
            return self._source.get(name)

        if name in self.netcdf_file.keys():
            return self.netcdf_file.get(name)

        return None

    def attr(self, var_name, attr_name):
        v = self.get(var_name)
        if v is None:
            return ''
        try:
            return v.attrs.get(attr_name)
        except AttributeError:
            pass
        return ''

    def __getitem__(self, name):
        # meta is special
        if name == 'meta':
            return self.meta

        if name in self._source:
            return self._source[name]

        if name in self.netcdf_file.keys():
            return self.netcdf_file.get(name)



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

KCU__IN_SUBSETS = {
    '1D': [0, 1, 2, 3, 4, 5],
    '1D2D': [51, 52, 53, 54, 55, 56, 57, 58],
    '2D_OPEN_WATER': [100, 101, 200, 300, 400, 500],
    'LONG_CRESTED_STRUCTURES': [3],
    'SHORT_CRESTED_STRUCTURES': [4],
    '2D_OPEN_WATER_OBSTACLES': [101],
    '2D_VERTICAL_INFILTRATION': [150],
    '2D_GROUNDWATER': [-150],
    'ACTIVE_BREACH': [56],
    'POTENTIAL_BREACH': [55]
}

KCU__IN_SUBSETS['1D_ALL'] = KCU__IN_SUBSETS['1D'] + KCU__IN_SUBSETS['1D2D']
KCU__IN_SUBSETS['GROUNDWATER_ALL'] = (
    KCU__IN_SUBSETS['2D_GROUNDWATER'] + KCU__IN_SUBSETS[
        '2D_VERTICAL_INFILTRATION']
)
KCU__IN_SUBSETS['2D_ALL'] = (
    KCU__IN_SUBSETS['GROUNDWATER_ALL'] +
    KCU__IN_SUBSETS['2D_OPEN_WATER_OBSTACLES'] +
    KCU__IN_SUBSETS['2D_OPEN_WATER']
)



from itertools import count
import os
import shutil
import sys

import h5py

if sys.version_info >= (3, 3):  # noqa
    from unittest import mock
else:
    import mock
import pytest
import six

from threedigrid.admin.constants import TYPE_CODE_MAP
from threedigrid.admin.gridresultadmin import GridH5ResultAdmin
from threedigrid.admin.idmapper import IdMapper


test_file_dir = os.path.join(os.path.dirname(__file__), "test_files")

grid_bck = os.path.join(test_file_dir, "gridadmin.bck")

# the testfile is a copy of the v2_bergermeer gridadmin file
result_file = os.path.join(test_file_dir, "results_3di.nc")
grid_file = os.path.join(test_file_dir, "gridadmin.h5")


NODE_LENGTH = 10


@pytest.fixture()
def gr():
    gr = GridH5ResultAdmin(grid_file, result_file)
    yield gr
    gr.close()


def simple_id_map(node_length=NODE_LENGTH):
    """Return a simple mapping

    NODE_LENGTH nodes are created for each TYPE_CODE, each mapping to an ever
    increasing sequence number"""
    sequence_id_generator = count(start=1, step=1)
    id_map = {}
    for code in TYPE_CODE_MAP.values():
        pk_generator = range(1, 1 + node_length)
        id_map[code] = dict(zip(pk_generator, sequence_id_generator))
    return id_map


@pytest.fixture
def empty_hdf5_file(tmpdir):
    """Create an empty hdf5 file"""
    hdf5_file_name = six.text_type(tmpdir.join("gridadmin.h5"))
    with h5py.File(hdf5_file_name, "w") as h5py_file:
        yield h5py_file


@pytest.fixture
def h5py_file(tmpdir):
    """Copy and open an unprepared h5py file, based on gridadmin.bck"""
    tmp_file = six.text_type(tmpdir.join("gridadmin.h5"))
    shutil.copyfile(grid_bck, tmp_file)
    with h5py.File(tmp_file, "r+") as h5py_file:
        yield h5py_file


@pytest.fixture
def threedi_datasource():
    """Mocks a threedi_datasource"""
    m = mock.MagicMock()
    return m


@pytest.fixture
@mock.patch("threedigrid.admin.idmapper.get_id_map")
def h5py_file_mapper(mocked_id_map, h5py_file, threedi_datasource):
    """Unprepared h5py file with a prepared id_mapper"""
    id_map = simple_id_map()
    mocked_id_map.return_value = id_map
    IdMapper.prepare_mapper(h5py_file, threedi_datasource)
    return h5py_file, threedi_datasource



import numpy as np
import pytest

from threedigrid.admin import prepare_utils
from threedigrid.admin.prepare_utils import add_or_update_datasets


@pytest.fixture(
    params=[
        [b""],
        [b"foo", b"bar"],
        [1, 2, 3],
        [-1.3, 0.0, 3.4],
        [b"a_very_long_string_" * 100],
    ]
)
def np_array_dict(request):
    data = np.array(request.param)
    pk = range(len(data))
    numpy_array_dict = {"pk": np.array(pk), "mydataset": data}

    # set the DEFAULT_NULL_VALUE based on the type of params.
    if type(request.param[0]) == bytes:
        prepare_utils.DEFAULT_NULL_VALUE = b""
    else:
        prepare_utils.DEFAULT_NULL_VALUE = -9999
    return numpy_array_dict


def test_add_or_update_datasets_new_data(empty_hdf5_file, np_array_dict):
    """Add some new data to an hdf5 file"""
    h5py_group = empty_hdf5_file.create_group("mygroup")
    field_names = np_array_dict.keys()
    content_pk = pk = np_array_dict["pk"]
    add_or_update_datasets(
        h5py_group, np_array_dict, field_names, pk, content_pk
    )
    assert list(h5py_group.keys()) == ["mydataset"]
    assert h5py_group["mydataset"].dtype == np_array_dict["mydataset"].dtype
    assert all(h5py_group["mydataset"][:] == np_array_dict["mydataset"])


def test_add_or_update_datasets_existing_data(empty_hdf5_file, np_array_dict):
    """Update existing data"""
    h5py_group = empty_hdf5_file.create_group("mygroup")
    h5py_group.create_dataset(
        name="mydataset", data=np_array_dict["mydataset"]
    )
    field_names = np_array_dict.keys()
    content_pk = pk = np_array_dict["pk"]
    add_or_update_datasets(
        h5py_group, np_array_dict, field_names, pk, content_pk
    )
    assert list(h5py_group.keys()) == ["mydataset"]
    assert h5py_group["mydataset"].dtype == np_array_dict["mydataset"].dtype
    assert all(h5py_group["mydataset"][:] == np_array_dict["mydataset"])



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import

import logging

import numpy as np
from threedigrid.admin.utils import PKMapper
from threedigrid.admin.prepare_utils import db_objects_to_numpy_array_dict

logger = logging.getLogger(__name__)


class PreparePumps:

    @staticmethod
    def prepare_datasource(h5py_file, threedi_datasource, pump_group):
        node_group = h5py_file['nodes']
        datasource = pump_group

        # Step 1: Load data from threedi_datasource
        pumpstations_field_names = [
            'pk', 'display_name', 'type', 'start_level',
            'lower_stop_level', 'capacity',
            'connection_node_start_pk', 'connection_node_end_pk',
            'zoom_category'
        ]

        pumpstations_numpy_array_dict = db_objects_to_numpy_array_dict(
            threedi_datasource.v2_pumpstations, pumpstations_field_names)

        ids = np.array([False])
        for field_name in pumpstations_field_names:
            data = pumpstations_numpy_array_dict[field_name]
            dataset_name = field_name

            if dataset_name == 'pk':
                dataset_name = 'content_pk'
                ids = np.arange(1, data.size + 1)

            # insert trash element
            data = np.insert(data.copy(), 0, 0, axis=len(data.shape) - 2)

            datasource.set(dataset_name, data)

        if np.any(ids):
            data = np.insert(ids.copy(), 0, 0, axis=len(data.shape) - 2)
            datasource.set('id', data)

        # Step 2: Postprocessing data

        # Connection nodes are nodes with content_pk != 0
        connection_nodes_mask = node_group['content_pk'].value != 0
        connection_nodes_content_pks = node_group[
            'content_pk'].value[connection_nodes_mask]
        connection_nodes_content_id = node_group['id'].value[
            connection_nodes_mask]
        connection_nodes_content_coordinates = node_group['coordinates'].value[
            :, connection_nodes_mask]

        connection_nodes_initial_waterlevel = node_group[
            'initial_waterlevel'].value[connection_nodes_mask]

        cn1_mapper = PKMapper(
            connection_nodes_content_pks,
            datasource['connection_node_start_pk'])
        cn2_mapper = PKMapper(
            connection_nodes_content_pks,
            datasource['connection_node_end_pk'])

        if 'node1_id' not in list(datasource.keys()):
            node1_id = cn1_mapper.apply_on(
                connection_nodes_content_id, -9999)
            datasource.set('node1_id', node1_id)

        if 'node2_id' not in list(datasource.keys()):
            node2_id = cn2_mapper.apply_on(
                connection_nodes_content_id, -9999)
            datasource.set('node2_id', node2_id)

        if 'bottom_level' not in list(datasource.keys()):
            bottom_level = cn1_mapper.apply_on(
                connection_nodes_initial_waterlevel, 0)
            datasource.set('bottom_level', bottom_level)

        if 'node_coordinates' not in list(datasource.keys()):
            node1_coordinates = cn1_mapper.apply_on(
                connection_nodes_content_coordinates, -9999)
            node2_coordinates = cn2_mapper.apply_on(
                connection_nodes_content_coordinates, -9999)

            datasource.set(
                'node_coordinates',
                np.vstack((node1_coordinates, node2_coordinates)))

        if 'coordinates' not in list(datasource.keys()):
            # Set the coordinates based on the centroid of node_coordinates
            # if both set and else the one that is set.
            node_coordinates = datasource['node_coordinates'].value
            node1_coords = node_coordinates[0:2, :]
            node2_coords = node_coordinates[2:4, :]
            node1_id = datasource['node1_id'].value
            node2_id = datasource['node2_id'].value
            node1_not_set = node1_id == -9999
            node2_not_set = node2_id == -9999

            # Replace empty values with coords from other node
            node2_coords[:, node2_not_set] = node1_coords[:, node2_not_set]
            node1_coords[:, node1_not_set] = node2_coords[:, node1_not_set]

            # Calculate centroids
            datasource.set(
                'coordinates',
                (node2_coords + node1_coords) / 2.0)



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
"""
"""

from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
from threedigrid.orm.base.timeseries_mixin import ResultMixin
from threedigrid.orm.base.timeseries_mixin import AggregateResultMixin

from threedigrid.orm.base.options import ModelMeta
import six


class PumpsResultsMixin(ResultMixin):

    class Meta(six.with_metaclass(ModelMeta)):
        field_attrs = ['units', 'long_name', 'standard_name']

        composite_fields = {'q_pump': ['Mesh1D_q_pump'],
                            '_mesh_id': ['Mesh1DPump_id']}

    def __init__(self, **kwargs):
        """Instantiate a breach with netcdf results.

        Variables stored in the netcdf and related to breaches are dynamically
        added as attributes as TimeSeriesArrayField.

        :param kwargs:
        """
        super(PumpsResultsMixin, self).__init__(**kwargs)


class PumpsAggregateResultsMixin(AggregateResultMixin):

    class Meta(six.with_metaclass(ModelMeta)):
        field_attrs = ['units', 'long_name']

        model_attr = 'timestamp'

        # ModelMeta will combine base_composition and composition_vars
        # to composite_fields attribute
        base_composition = {'q_pump': ['Mesh1D_q_pump'],
                            '_mesh_id': ['Mesh1DPump_id']}
        composition_vars = {
            'q_pump':
                ['min', 'max', 'avg', 'cum', 'cum_positive', 'cum_negative'],
        }

    def __init__(self, **kwargs):
        """Instantiate a line with netcdf results.

        Variables stored in the netcdf and related to lines are dynamically
        added as attributes as TimeSeriesArrayField.

        :param netcdf_keys: list of netcdf variables
        :param kwargs:
        """
        super(PumpsAggregateResultsMixin, self).__init__(**kwargs)



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import logging
from collections import defaultdict

import h5py

from threedigrid.admin.breaches.models import Breaches
from threedigrid.admin.breaches.timeseries_mixin import (
    BreachesAggregateResultsMixin, BreachesResultsMixin)
from threedigrid.admin.constants import DEFAULT_CHUNK_TIMESERIES
from threedigrid.admin.gridadmin import GridH5Admin
from threedigrid.admin.h5py_datasource import H5pyResultGroup
from threedigrid.admin.h5py_swmr import H5SwmrFile
from threedigrid.admin.lines.models import Lines
from threedigrid.admin.lines.timeseries_mixin import LinesAggregateResultsMixin
from threedigrid.admin.lines.timeseries_mixin import LinesResultsMixin
from threedigrid.admin.nodes.models import Nodes
from threedigrid.admin.nodes.timeseries_mixin import NodesAggregateResultsMixin
from threedigrid.admin.nodes.timeseries_mixin import NodesResultsMixin
from threedigrid.admin.pumps.models import Pumps
from threedigrid.admin.pumps.timeseries_mixin import PumpsAggregateResultsMixin
from threedigrid.admin.pumps.timeseries_mixin import PumpsResultsMixin
from threedigrid.orm.models import Model

logger = logging.getLogger(__name__)

try:
    import asyncio_rpc # noqa
    asyncio_rpc_support = True
except ImportError:
    asyncio_rpc_support = False


class GridH5ResultAdmin(GridH5Admin):
    """
    Admin interface for threedicore result queries.
    """

    def __init__(
            self, h5_file_path, netcdf_file_path, file_modus='r', swmr=False):
        """

        :param h5_file_path: path to the hdf5 gridadmin file
        :param netcdf_file_path: path to the netcdf result file (usually
            called subgrid_map.nc)
        :param file_modus: modus in which to open the files
        """
        self._field_model_dict = defaultdict(list)
        self._netcdf_file_path = netcdf_file_path
        super(GridH5ResultAdmin, self).__init__(h5_file_path, file_modus)

        self.result_datasource_class = H5pyResultGroup

        if h5_file_path.startswith('rpc://'):
            if not asyncio_rpc_support:
                raise Exception(
                    "Please reinstall this package with threedigrid[rpc]")

            from threedigrid.admin.rpc_datasource import (
                H5RPCResultGroup, RPCFile)
            self.netcdf_file = RPCFile(h5_file_path, file_modus)
            self.result_datasource_class = H5RPCResultGroup
        else:
            if swmr:
                self.netcdf_file = H5SwmrFile(
                    netcdf_file_path, file_modus)
            else:
                self.netcdf_file = h5py.File(
                    netcdf_file_path, file_modus)
            self.version_check()
        self.set_timeseries_chunk_size(DEFAULT_CHUNK_TIMESERIES.stop)

    def set_timeseries_chunk_size(self, new_chunk_size):
        """
        overwrite the default chunk size for timeseries queries.
        :param new_chunk_size <int>: new chunk size for
            timeseries queries
        :raises ValueError when the given value is less than 1
        """
        _chunk_size = int(new_chunk_size)
        if _chunk_size < 1:
            raise ValueError('Chunk size must be greater than 0')
        self._timeseries_chunk_size = slice(0, _chunk_size)
        logger.info(
            'New chunk for timeseries size has been set to %d',
            new_chunk_size
        )
        self._grid_kwargs.update(
            {'timeseries_chunk_size': self._timeseries_chunk_size}
        )

    @property
    def timeseries_chunk_size(self):
        return self._timeseries_chunk_size.stop

    @property
    def time_units(self):
        return self.netcdf_file['time'].attrs.get('units')

    @property
    def lines(self):
        return Lines(
            self.result_datasource_class(
                self.h5py_file, 'lines', self.netcdf_file),
            **dict(self._grid_kwargs, **{'mixin': LinesResultsMixin}))

    @property
    def nodes(self):
        return Nodes(
            self.result_datasource_class(
                self.h5py_file, 'nodes', self.netcdf_file),
            **dict(self._grid_kwargs, **{'mixin': NodesResultsMixin}))

    @property
    def breaches(self):
        if not self.has_breaches:
            logger.info('Threedimodel has no breaches')
            return
        return Breaches(
            self.result_datasource_class(
                self.h5py_file, 'breaches', self.netcdf_file),
            **dict(self._grid_kwargs, **{'mixin': BreachesResultsMixin}))

    @property
    def pumps(self):
        if not self.has_pumpstations:
            logger.info('Threedimodel has no pumps')
            return
        return Pumps(
            self.result_datasource_class(
                self.h5py_file, 'pumps', self.netcdf_file),
            **dict(self._grid_kwargs, **{'mixin': PumpsResultsMixin}))

    def version_check(self):
        """
        compare versions of grid admin and grid results file.
        Issues a warning if they differ
        """

        if self.threedicore_result_version != self.threedicore_version:
            logger.warning(
                '[!] threedicore version differ! \n'
                'Version result file has been created with: %s\n'
                'Version gridadmin file has been created with: %s',
                self.threedicore_result_version, self.threedicore_version
            )

    @property
    def _field_model_map(self):
        """
        :return: a dict of {<field name>: [model name, ...]}
        """
        if self._field_model_dict:
            return self._field_model_dict

        model_names = set()
        for attr_name in dir(self):
            # skip private attrs
            if any([attr_name.startswith('__'),
                    attr_name.startswith('_')]):
                continue
            try:
                attr = getattr(self, attr_name)
            except AttributeError:
                logger.warning("Attribute: '{}' does not "
                               "exist in h5py_file.".format(attr_name))
                continue
            if not issubclass(type(attr), Model):
                continue
            model_names.add(attr_name)

        for model_name in model_names:
            for x in getattr(self, model_name)._field_names:
                self._field_model_dict[x].append(model_name)
        return self._field_model_dict

    def get_model_instance_by_field_name(self, field_name):
        """
        :param field_name: name of a models field
        :return: instance of the model the field belongs to
        :raises IndexError if the field name is not unique across models
        """
        model_name = self._field_model_map.get(field_name)
        if not model_name or len(model_name) != 1:
            raise IndexError(
                'Ambiguous result. Field name {} yields {} model(s)'.format(
                    field_name, len(model_name) if model_name else 0)
            )
        return getattr(self, model_name[0])

    @property
    def threedicore_result_version(self):
        """
        :return: version of the grid result file (if found), an empty
        string otherwise
        """
        try:
            return self.netcdf_file.attrs.get('threedicore_version')
        except AttributeError:
            logger.error(
                'Attribute threedicore_version could not be found in result file')  # noqa
        return ''

    def close(self):
        super(GridH5ResultAdmin, self).close()
        self.netcdf_file.close()


class GridH5AggregateResultAdmin(GridH5ResultAdmin):
    """
    Admin interface for threedicore result queries.
    """
    def __init__(self, h5_file_path, netcdf_file_path, file_modus='r'):
        """

        :param h5_file_path: path to the hdf5 gridadmin file
        :param netcdf_file_path: path to the netcdf result file (usually
            called subgrid_map.nc)
        :param file_modus: modus in which to open the files
        """
        super(GridH5AggregateResultAdmin, self).__init__(
            h5_file_path, netcdf_file_path, file_modus)

    @property
    def lines(self):
        model_name = 'lines'
        return Lines(
            H5pyResultGroup(self.h5py_file, model_name, self.netcdf_file),
            **dict(self._grid_kwargs, **{'mixin': LinesAggregateResultsMixin})
        )

    @property
    def nodes(self):
        return Nodes(
            H5pyResultGroup(self.h5py_file, 'nodes', self.netcdf_file),
            **dict(self._grid_kwargs, **{'mixin': NodesAggregateResultsMixin}))

    @property
    def breaches(self):
        if not self.has_breaches:
            logger.info('Threedimodel has no breaches')
            return
        model_name = 'breaches'
        return Breaches(
            H5pyResultGroup(self.h5py_file, model_name, self.netcdf_file),
            **dict(self._grid_kwargs,
                   **{'mixin': BreachesAggregateResultsMixin})
        )

    @property
    def pumps(self):
        if not self.has_pumpstations:
            logger.info('Threedimodel has no pumps')
            return
        model_name = 'pumps'
        return Pumps(
            H5pyResultGroup(self.h5py_file, model_name, self.netcdf_file),
            **dict(self._grid_kwargs, **{'mixin': PumpsAggregateResultsMixin}))

    @property
    def time_units(self):
        logger.info(
            'Time units are not defined globally for aggregated results')
        return None



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import json
import os
import unittest
import tempfile
import shutil

from osgeo import ogr

from threedigrid.admin import constants
from threedigrid.admin.gridadmin import GridH5Admin
from threedigrid.admin.prepare import GridAdminH5Export

from threedigrid.admin.lines.exporters import LinesOgrExporter


test_file_dir = os.path.join(os.getcwd(), "tests/test_files")

# the testfile is a copy of the v2_bergermeer gridadmin file
grid_admin_h5_file = os.path.join(test_file_dir, "gridadmin.h5")


class ExporterTestShp(unittest.TestCase):
    def setUp(self):
        self.parser = GridH5Admin(grid_admin_h5_file)
        d = tempfile.mkdtemp()
        self.f_shp = os.path.join(d, "exporter_test_lines.shp")

    def tearDown(self):
        shutil.rmtree(os.path.dirname(self.f_shp))

    def test_export_by_extension(self):
        line_2d_open_water_wgs84 = self.parser.lines.subset(
            "2D_OPEN_WATER"
        ).reproject_to("4326")
        exporter = LinesOgrExporter(line_2d_open_water_wgs84)
        exporter.save(self.f_shp, line_2d_open_water_wgs84.data, "4326")
        self.assertTrue(os.path.exists(self.f_shp))
        s = ogr.Open(self.f_shp)
        layer = s.GetLayer()
        self.assertEqual(
            layer.GetFeatureCount(), line_2d_open_water_wgs84.id.size
        )


class ExporterTestGpkg(unittest.TestCase):
    def setUp(self):
        self.parser = GridH5Admin(grid_admin_h5_file)
        d = tempfile.mkdtemp()
        self.f_gpkg = os.path.join(d, "exporter_test_lines.gpkg")

    def tearDown(self):
        shutil.rmtree(os.path.dirname(self.f_gpkg))

    def test_export_by_extension(self):
        line_2d_open_water_wgs84 = self.parser.lines.subset(
            "2D_OPEN_WATER"
        ).reproject_to("4326")
        exporter = LinesOgrExporter(line_2d_open_water_wgs84)
        exporter.save(self.f_gpkg, line_2d_open_water_wgs84.data, "4326")
        self.assertTrue(os.path.exists(self.f_gpkg))
        s = ogr.Open(self.f_gpkg)
        layer = s.GetLayer()
        self.assertEqual(
            layer.GetFeatureCount(), line_2d_open_water_wgs84.id.size
        )


class ExporterTestGeojson(unittest.TestCase):
    def setUp(self):
        self.parser = GridH5Admin(grid_admin_h5_file)
        d = tempfile.mkdtemp()
        self.f_geojson = os.path.join(d, "exporter_test_lines.json")

    def tearDown(self):
        shutil.rmtree(os.path.dirname(self.f_geojson))

    def test_export_by_extension(self):
        # .json extension automatically uses geojson exporter
        line_2d_open_water_wgs84 = self.parser.lines.subset(
            "2D_OPEN_WATER"
        ).reproject_to("4326")
        exporter = LinesOgrExporter(line_2d_open_water_wgs84)
        exporter.save(self.f_geojson, line_2d_open_water_wgs84.data, "4326")
        self.assertTrue(os.path.exists(self.f_geojson))
        s = ogr.Open(self.f_geojson)
        layer = s.GetLayer()
        self.assertEqual(
            layer.GetFeatureCount(), line_2d_open_water_wgs84.id.size
        )

    def test_export_specify_fields_as_dict(self):
        self.parser.lines.pipes.filter(id=27449).to_geojson(
            self.f_geojson,
            use_ogr=False,
            fields={'display_name': 'str', 'calculation_type': 'str'}
        )
        with open(self.f_geojson) as file:
            data = json.load(file)
            self.assertEqual(len(data['features'][0]['properties']), 3)
            self.assertTrue(
                'display_name' in data['features'][0]['properties']
            )
            self.assertTrue(
                'calculation_type' in data['features'][0]['properties']
            )
            self.assertTrue(
                'model_type' in data['features'][0]['properties']
            )

    def test_export_specify_fields_as_list(self):
        self.parser.lines.pipes.filter(id=27449).to_geojson(
            self.f_geojson,
            use_ogr=False,
            fields=['display_name', 'calculation_type']
        )
        with open(self.f_geojson) as file:
            data = json.load(file)
            self.assertEqual(len(data['features'][0]['properties']), 3)
            self.assertTrue(
                'display_name' in data['features'][0]['properties']
            )
            self.assertTrue(
                'calculation_type' in data['features'][0]['properties']
            )
            self.assertTrue(
                'model_type' in data['features'][0]['properties']
            )

    def test_export_specify_nested_fields(self):
        self.parser.lines.pipes.filter(id=27449).to_geojson(
            self.f_geojson,
            use_ogr=False,
            fields=[
                {
                    'first': [
                        'display_name',
                        {'second': ['calculation_type']}
                    ]
                }
            ]
        )
        with open(self.f_geojson) as file:
            data = json.load(file)
            self.assertEqual(len(data['features'][0]['properties']), 2)
            self.assertTrue(
                'first' in data['features'][0]['properties']
            )
            self.assertTrue(
                'model_type' in data['features'][0]['properties']
            )
            self.assertTrue(
                'display_name' in data['features'][0]['properties']['first']
            )
            self.assertTrue(
                'calculation_type' in
                data['features'][0]['properties']['first']['second']
            )

    def test_export_filter(self):
        self.parser.lines.pipes.filter(id=27449).to_geojson(
            self.f_geojson, use_ogr=False
        )
        with open(self.f_geojson) as file:
            data = json.load(file)
            self.assertEqual(len(data['features']), 1)

    def test_export_reproject(self):
        self.parser.lines.reproject_to("4326").to_geojson(
            self.f_geojson, use_ogr=False
        )
        with open(self.f_geojson) as file:
            data = json.load(file)
            p1, _ = data['features'][1]['geometry']['coordinates']
            x, y = p1
            # Should be somewhere in the Netherlands
            self.assertTrue(4.185791 < x < 5.663452)
            self.assertTrue(52.109879 < y < 52.912215)


class GridadminH5ExportTest(unittest.TestCase):
    def setUp(self):
        self.d = tempfile.mkdtemp()
        self.exporter = GridAdminH5Export(grid_admin_h5_file)
        self.exporter._dest = self.d

    def tearDown(self):
        shutil.rmtree(self.d)

    def test_export_2d_groundwater(self):
        self.exporter.export_2d_groundwater_lines()
        result = os.path.join(
            self.d, constants.GROUNDWATER_LINES + self.exporter._extension
        )
        self.assertTrue(os.path.exists(result))

    def test_export_2d_openwater_lines(self):
        self.exporter.export_2d_openwater_lines()
        result = os.path.join(
            self.d, constants.OPEN_WATER_LINES + self.exporter._extension
        )
        self.assertTrue(os.path.exists(result))

    def test_export_2d_vertical_infiltration_lines(self):
        self.exporter.export_2d_vertical_infiltration_lines()
        result = os.path.join(
            self.d,
            constants.VERTICAL_INFILTRATION_LINES + self.exporter._extension
        )
        self.assertTrue(os.path.exists(result))

    def test_export_levees(self):
        self.exporter.export_levees()
        result = os.path.join(
            self.d, constants.LEVEES + self.exporter._extension
        )
        self.assertTrue(os.path.exists(result))



import os
import sys

if sys.version_info >= (3, 3):  # noqa
    from unittest import mock
else:
    import mock

from threedigrid.admin.idmapper import IdMapper
from threedigrid.admin.prepare import GridAdminH5Prepare
from threedigrid.admin.prepare import is_prepared
from .conftest import simple_id_map


test_file_dir = os.path.join(os.path.dirname(__file__), "test_files")

grid_bck = os.path.join(test_file_dir, "gridadmin.bck")


def test_prepare_onedee_lines(h5py_file, threedi_datasource):
    IdMapper.prepare_mapper(h5py_file, threedi_datasource)
    GridAdminH5Prepare.prepare_onedee_lines(h5py_file, threedi_datasource)
    # check some special fixed types
    assert h5py_file["lines"]["code"].dtype == "S32"
    assert h5py_file["lines"]["display_name"].dtype == "S64"
    assert is_prepared(h5py_file, "lines", "lines_prepared")


def test_prepare_lines(h5py_file, threedi_datasource):
    IdMapper.prepare_mapper(h5py_file, threedi_datasource)
    GridAdminH5Prepare.prepare_lines(h5py_file, threedi_datasource)
    assert is_prepared(h5py_file, "lines", "lines_prepared")


def test_prepare_onedee_nodes(h5py_file_mapper):
    h5py_file, threedi_datasource = h5py_file_mapper
    GridAdminH5Prepare.prepare_onedee_nodes(h5py_file, threedi_datasource)
    assert is_prepared(h5py_file, "nodes", "connectionnodes_prepared")
    assert is_prepared(h5py_file, "nodes", "manholes_prepared")
    connection_nodes_field_names = {"initial_waterlevel", "storage_area"}
    assert connection_nodes_field_names.issubset(h5py_file["nodes"].keys())


def test_prepare_nodes(h5py_file_mapper):
    h5py_file, threedi_datasource = h5py_file_mapper
    GridAdminH5Prepare.prepare_nodes(h5py_file, threedi_datasource)
    assert is_prepared(h5py_file, "nodes", "prepared")
    assert {"id", "content_pk", "seq_id", "coordinates"}.issubset(
        h5py_file["nodes"].keys()
    )


def test_prepare_pumps(h5py_file_mapper):
    h5py_file, threedi_datasource = h5py_file_mapper
    # need to prepare onedee_nodes before we can prepare pumps
    GridAdminH5Prepare.prepare_onedee_nodes(h5py_file, threedi_datasource)
    GridAdminH5Prepare.prepare_pumps(h5py_file, threedi_datasource)
    assert is_prepared(h5py_file, "pumps", "prepared")
    pumpstations_field_names = {
        "display_name",
        "start_level",
        "lower_stop_level",
        "capacity",
        "connection_node_start_pk",
        "connection_node_end_pk",
        "zoom_category",
    }
    assert pumpstations_field_names.issubset(h5py_file["pumps"].keys())


def test_prepare_levees(h5py_file_mapper):
    h5py_file, threedi_datasource = h5py_file_mapper
    GridAdminH5Prepare.prepare_levees(h5py_file, threedi_datasource)
    assert is_prepared(h5py_file, "levees", "prepared")
    levees_field_names = {"coords", "crest_level", "max_breach_depth", "id"}
    assert levees_field_names.issubset(h5py_file["levees"].keys())


@mock.patch("threedigrid.admin.idmapper.get_id_map")
def test_prepare_breaches(mocked_id_map, h5py_file, threedi_datasource):
    id_map = simple_id_map()
    id_map[13] = {0: 1, 1: 1, 2: 1}
    mocked_id_map.return_value = id_map
    IdMapper.prepare_mapper(h5py_file, threedi_datasource)
    GridAdminH5Prepare.prepare_breaches(h5py_file, threedi_datasource)
    assert is_prepared(h5py_file, "breaches", "prepared")
    breaches_field_names = {
        "id",
        "seq_ids",
        "content_pk",
        "kcu",
        "coordinates",
    }
    assert breaches_field_names.issubset(h5py_file["breaches"].keys())


@mock.patch("threedigrid.admin.idmapper.get_id_map")
def test_prepare(mocked_id_map, h5py_file, threedi_datasource):
    # breaches fix
    id_map = simple_id_map()
    id_map[13] = {0: 1}
    mocked_id_map.return_value = id_map
    GridAdminH5Prepare.prepare(h5py_file, threedi_datasource)
    assert is_prepared(h5py_file, "lines", "lines_prepared")
    assert is_prepared(h5py_file, "nodes", "prepared")
    assert is_prepared(h5py_file, "pumps", "prepared")
    assert is_prepared(h5py_file, "levees", "prepared")
    assert is_prepared(h5py_file, "breaches", "prepared")



from __future__ import absolute_import
import unittest

import numpy as np
from shapely.geometry import Polygon, Point, LineString

from threedigrid.orm.base.filters import BaseFilter
from threedigrid.orm.base.filters import BaseCompareFilter
from threedigrid.orm.base.filters import EqualsFilter
from threedigrid.orm.base.filters import get_filter
from threedigrid.orm.base.filters import InFilter
from threedigrid.orm.filters import PointFilter, GeometryIntersectionFilter
from threedigrid.orm.filters import FILTER_MAP

from threedigrid.orm.base.fields import ArrayField
from threedigrid.orm.fields import PolygonArrayField, BboxArrayField, \
    LineArrayField, PointArrayField, MultiLineArrayField


class BasicFilterTests(unittest.TestCase):
    def setUp(self):
        self.field = ArrayField()

    def test_base_filter(self):
        b = BaseFilter()
        with self.assertRaises(NotImplementedError):
            b.filter("foo")

    def test_repr(self):
        f = BaseCompareFilter("in", self.field, [1, 2, 3])
        self.assertTrue(repr(f))

    def test_get_filter_eq(self):
        """Test that we get a default EqualsFilter without keys."""
        f = get_filter(["content_type"], self.field, 3)
        self.assertIsInstance(f, EqualsFilter)

        # also works when 'eq' is explicitly given
        f = get_filter(["content_type", "eq"], self.field, 3)
        self.assertIsInstance(f, EqualsFilter)

    def test_eq_filter(self):
        f = EqualsFilter("content_type", self.field, np.array([1, 2]))
        filtered = f.filter({"content_type": np.array([1, 2])})
        self.assertTrue(filtered.all())

        filtered = f.filter({"content_type": np.array([1, 3])})
        self.assertFalse(filtered.all())

    def test_get_filter_in(self):
        f = get_filter(["content_pk", "in"], self.field, [2])
        self.assertIsInstance(f, InFilter)

    def test_get_filter_error(self):
        """get_filter fails with malformed splitted keys."""
        with self.assertRaises(ValueError):
            get_filter(["content_pk", "in", "foo"], self.field, [2])

        with self.assertRaises(ValueError):
            get_filter([], self.field, [2])

    def test_get_nonexisting_filter(self):
        with self.assertRaises(ValueError):
            get_filter(["content_type", "not_a_filter"], self.field, 3)


class PointFilterTests(unittest.TestCase):
    def setUp(self):
        self.field = PolygonArrayField()

    def test_get_point_filter(self):
        f = get_filter(
            ["cell_coords", "contains_point"],
            self.field,
            [1, 5],
            filter_map=FILTER_MAP,
        )
        self.assertIsInstance(f, PointFilter)

    def test_point_filter(self):
        f = PointFilter("coord_cells", self.field, np.array([1, 5]))
        filtered = f.filter(
            {"coord_cells": np.array([[0, 2], [0, 2], [6, 6], [10, 11]])}
        )
        self.assertTrue(filtered[0])
        self.assertFalse(filtered[1])


class GeometryIntersectionFilterTest(unittest.TestCase):

    def setUp(self):
        self.geometry = Polygon(
            [[0, 0], [0, 1], [1, 1], [0, 0]]
        )
        self.field = BboxArrayField()

    def test_get_intersect_geometry_filter(self):
        f = get_filter(
            ['cell_coords', 'intersects_geometry'],
            BboxArrayField(),
            self.geometry,
            filter_map=FILTER_MAP
        )
        self.assertIsInstance(f, GeometryIntersectionFilter)

    def test_intersection_geometry_filter_bbox_array(self):
        f = GeometryIntersectionFilter(
            'cell_coords', BboxArrayField(), self.geometry
        )
        filtered = f.filter(
            {"cell_coords": np.array([
                [0, 0, 1, 1], [2, 2, 3, 3]
            ]).T}
        )
        self.assertTrue(filtered[0])
        self.assertFalse(filtered[1])

    def test_intersection_geometry_filter_bbox_array_line_bbox(self):
        # the BBOX of the line intersects all cells, but the line intersects
        # only 3 cells.
        line = LineString([(0.5, 0.5), (1.5, 0.5), (1.5, 1.5)])
        values = np.array([[0, 0, 1, 1],
                           [0, 1, 0, 1],
                           [1, 1, 2, 2],
                           [1, 2, 1, 2]])
        f = GeometryIntersectionFilter(
            'cell_coords', BboxArrayField(), line
        )
        filtered = f.filter({"cell_coords": values})
        self.assertTrue(filtered[0])
        self.assertTrue(filtered[2])
        self.assertTrue(filtered[3])
        self.assertFalse(filtered[1])

    def test_intersection_geometry_filter_line_array(self):
        f = GeometryIntersectionFilter(
            'line_coords', LineArrayField(), self.geometry
        )
        filtered = f.filter(
            {"line_coords": np.array([
                [0, 0, 1, 1], [2, 2, 3, 3]
            ]).T}
        )
        self.assertTrue(filtered[0])
        self.assertFalse(filtered[1])

    def test_intersection_geometry_filter_point_array(self):
        f = GeometryIntersectionFilter(
            'coordinates', PointArrayField(), self.geometry
        )
        filtered = f.filter(
            {"coordinates": np.array([
                [0.5, 2], [0.5, 2]
            ])}
        )
        self.assertTrue(filtered[0])
        self.assertFalse(filtered[1])

    def test_intersection_geometry_filter_polygon_array(self):
        f = GeometryIntersectionFilter(
            'coordinates', PolygonArrayField(), self.geometry
        )
        filtered = f.filter(
            {"coordinates": np.array(
                [  # + sign shows the split between x, y coords
                    np.array([0, 0, 2, 2] + [0, 2, 2, 0]),
                    np.array([5, 6, 6] + [5, 6, 5])
                ]
            )}
        )
        self.assertTrue(filtered[0])
        self.assertFalse(filtered[1])

    def test_intersection_geometry_filter_multiline_array(self):
        f = GeometryIntersectionFilter(
            'coords', MultiLineArrayField(), self.geometry
        )
        filtered = f.filter(
            {"coords": np.array(
                [  # + sign shows the split between x, y coords
                    np.array([0, 0.5, 2, 2] + [0, 0.5, 2, 0]),
                    np.array([5, 6, 6] + [5, 6, 5])
                ]
            )}
        )
        self.assertTrue(filtered[0])
        self.assertFalse(filtered[1])

    def test_intersection_geometry_filter_point_geom(self):
        geometry = Point(0.5, 0.5)
        f = GeometryIntersectionFilter(
            'coordinates', PointArrayField(), geometry
        )
        filtered = f.filter(
            {"coordinates": np.array([
                [0.5, 2], [0.5, 2]
            ])}
        )
        self.assertTrue(filtered[0])
        self.assertFalse(filtered[1])

    def test_intersection_geometry_filter_line_geom(self):
        geometry = LineString([(0.5, 0.5), (1.5, 1.5)])
        f = GeometryIntersectionFilter(
            'cell_coords', BboxArrayField(), geometry
        )
        filtered = f.filter(
            {"cell_coords": np.array([
                [0, 0, 1, 1], [2, 2, 3, 3]
            ]).T}
        )
        self.assertTrue(filtered[0])
        self.assertFalse(filtered[1])



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
import pytest

from threedigrid.admin.gridresultadmin import GridH5ResultAdmin
from threedigrid.orm.base.options import ModelMeta
import six

test_file_dir = os.path.join(os.getcwd(), "tests/test_files")

# the testfile is a copy of the v2_bergermeer gridadmin file
result_file = os.path.join(test_file_dir, "results_3di.nc")
grid_file = os.path.join(test_file_dir, "gridadmin.h5")


@pytest.fixture()
def gr():
    gr = GridH5ResultAdmin(grid_file, result_file)
    yield gr
    gr.close()


def test_combine_vars():
    a = {"a", "b"}
    b = {"c", "d"}
    result = ModelMeta.combine_vars(a, b, "_")
    assert set(result) == {"a_c", "a_d", "b_c", "b_d"}


def test_combine_vars_empty_set():
    a = {"a", "b"}
    b = {}
    result = ModelMeta.combine_vars(a, b, "_")
    assert result == []


def test_combine_vars_invalid_combinator():
    a = {"a", "b"}
    b = {"c"}
    with pytest.raises(TypeError):
        ModelMeta.combine_vars(a, b, 5)


BASE_COMPOSITE_FIELDS = {
    "au": ["Mesh2D_au", "Mesh1D_au"],
    "u1": ["Mesh2D_u1", "Mesh1D_u1"],
    "q": ["Mesh2D_q", "Mesh1D_q"],
    "_mesh_id": ["Mesh2DLine_id", "Mesh1DLine_id"],  # private
}


@pytest.fixture()
def mm_composite():
    class TestMetaComposition(object):
        class Meta(six.with_metaclass(ModelMeta)):
            composite_fields = {"au": ["Mesh2D_au", "Mesh1D_au"]}

    tmc = TestMetaComposition()
    yield tmc


@pytest.fixture()
def mm_base_composite():
    class TestMetaBase(object):
        class Meta(six.with_metaclass(ModelMeta)):
            base_composition = {"q": ["Mesh2D_q", "Mesh1D_q"]}

            composition_vars = {"q": ["cum", "cum_positive", "cum_negative"]}

    tmc = TestMetaBase()
    yield tmc


def test_meta_error():
    # must define a composite_fields dict
    with pytest.raises(AttributeError):

        class TestMetaComposition(object):
            class Meta(six.with_metaclass(ModelMeta)):
                pass

        TestMetaComposition()


def test_meta_composite(mm_composite):
    assert mm_composite.Meta.composite_fields == {
        "au": ["Mesh2D_au", "Mesh1D_au"]
    }


def test_meta_base_composition(mm_base_composite):
    assert {"q_cum", "q_cum_positive", "q_cum_negative"} == set(
        mm_base_composite.Meta.composite_fields.keys()
    )
    import itertools

    assert set(
        list(
            itertools.chain(
                *list(mm_base_composite.Meta.composite_fields.values())
            )
        )
    ) == {
        "Mesh2D_q_cum",
        "Mesh1D_q_cum",
        "Mesh2D_q_cum_positive",
        "Mesh1D_q_cum_positive",
        "Mesh2D_q_cum_negative",
        "Mesh1D_q_cum_negative",
    }



"""
The admin module hosts all functionality that is necessary to connect to the
grid administration file. That is, the data can be read, altered, serialized
and exported by the classes and methods implemented here.
"""



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import sys

import numpy as np
import pytest

from threedigrid.admin.constants import LONLAT_DIGITS
from threedigrid.admin.utils import PKMapper
from threedigrid.admin.utils import _get_storage_area
from threedigrid.geo_utils import transform_bbox
from threedigrid.numpy_utils import get_smallest_uint_dtype
from threedigrid import numpy_utils
from threedigrid.orm.base.utils import _flatten_dict_values


def test_create_np_lookup_index_for():
    a = np.arange(6)
    b = np.array([3, 4, 1, 2, 5, 0])
    lidx = numpy_utils.create_np_lookup_index_for(a, b)
    np.testing.assert_array_equal(lidx, np.array([5, 2, 3, 0, 1, 4]))


def test_pk_mapper():
    pk = np.array([1, 2, 3, 4, 5, 6])
    to_map = np.array([1.1, 2.1, 3.1, 4.1, 5.1, 6.1])
    content_pk = np.array([0, 0, 15, 3, 4, 0, 0, 1, 1, 15])
    expected = [0, 0, 0, 3.1, 4.1, 0, 0, 1.1, 1.1, 0]
    result = PKMapper(pk, content_pk).apply_on(to_map, 0)
    np.testing.assert_equal(result, expected)


def test_get_smallest_uint_dtype():
    r = get_smallest_uint_dtype(4)
    assert r == np.uint8
    rr = get_smallest_uint_dtype(400)
    assert rr == np.uint16
    rrr = get_smallest_uint_dtype(400000)
    assert rrr == np.uint32
    rrrr = get_smallest_uint_dtype(40000000000)
    assert rrrr == np.uint64


def test_get_smallest_uint_dtype_raises_value_error_neg_input():
    with pytest.raises(ValueError):
        get_smallest_uint_dtype(-1)


def test_get_smallest_uint_dtype_raises_value_error_exceeding_input():
    with pytest.raises(ValueError):
        get_smallest_uint_dtype(400000000000000000000000)


@pytest.mark.skipif(sys.version_info < (3, 0), reason="requires Python3")
def test_get_storage_area_bytes_to_float():
    a = np.array(["2.2"], np.bytes_)
    sa_a = _get_storage_area(a)
    assert sa_a == 2.2


@pytest.mark.skipif(sys.version_info < (3, 0), reason="requires Python3")
def test_get_storage_area_bytes_to_string():
    b = np.array([""], np.bytes_)
    sa_b = _get_storage_area(b)
    assert sa_b == "--"  # default response
    b = np.array(["0"], np.bytes_)
    sa_b = _get_storage_area(b)
    assert sa_b == "--"  # default response


@pytest.mark.skipif(sys.version_info > (3, 0), reason="requires Python2")
def test_get_storage_area_string_input():
    b = np.array([""], np.string_)
    sa_b = _get_storage_area(b)
    assert sa_b == "--"  # default response
    b = np.array(["0"], np.string_)
    sa_b = _get_storage_area(b)
    assert sa_b == "--"  # default response


@pytest.mark.skipif(sys.version_info > (3, 0), reason="requires Python2")
def test_get_storage_area_string_to_float():
    b = np.array(["0.5"], np.string_)
    sa_b = _get_storage_area(b)
    assert sa_b == 0.5


def test_get_storage_area_none_input():
    sa_c = _get_storage_area(None)
    assert sa_c == "--"


def test_flatten_dict_values_as_set():
    d = {"a": 1}
    v = _flatten_dict_values(d, as_set=True)
    assert v == {1}


def test_flatten_nested_dict_values_as_set():
    d = {"a": [1]}
    v = _flatten_dict_values(d, as_set=True)
    assert v == {1}


def test_flatten_dict_values_as_list():
    d = {"a": 1}
    v = _flatten_dict_values(d, as_set=False)
    assert v == [1]


def test_flatten_nested_dict_values_as_list():
    d = {"a": [1]}
    v = _flatten_dict_values(d, as_set=False)
    assert v == [1]


def test_transform_bbox():
    bbox = np.array([106314., 517472., 106474., 517632.])
    source_epsg = "28992"
    target_epsg = "4326"
    transformed_bbox = transform_bbox(bbox, source_epsg, target_epsg)
    expected_result = np.array(
        [4.66789993, 52.64256428,  4.67024022, 52.64401638]
    )
    assert transformed_bbox.shape == (4,)
    np.testing.assert_allclose(
        transformed_bbox, expected_result, 10**-LONLAT_DIGITS
    )


def test_transform_bbox_all_coords():
    bbox = np.array([106314., 517472., 106474., 517632.])
    source_epsg = "28992"
    target_epsg = "4326"
    transformed_bbox = transform_bbox(
        bbox, source_epsg, target_epsg, all_coords=True
    )
    assert transformed_bbox.shape == (8,)



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
import unittest
import json

from threedigrid.admin.gridadmin import GridH5Admin
from threedigrid.admin.lines.serializers import ChannelsGeoJsonSerializer
from threedigrid.admin.lines.serializers import WeirsGeoJsonSerializer
from threedigrid.admin.breaches.serializers import BreachesGeoJsonSerializer
from threedigrid.admin.nodes.serializers import ManholesGeoJsonSerializer
from threedigrid.admin.pumps.serializers import PumpsGeoJsonSerializer

test_file_dir = os.path.join(os.getcwd(), "tests/test_files")

# the testfile is a copy of the v2_bergermeer gridadmin file
grid_admin_h5_file = os.path.join(test_file_dir, "gridadmin.h5")


class GridAdminSerializerTest(unittest.TestCase):
    def setUp(self):
        self.parser = GridH5Admin(grid_admin_h5_file)

    def test_channel_serializer(self):
        channels_wgs84 = self.parser.lines.channels.subset(
            "1D_ALL"
        ).reproject_to("4326")
        channels_wgs84_geojson = ChannelsGeoJsonSerializer(channels_wgs84).data
        j = json.loads(channels_wgs84_geojson)
        feat = j["features"]
        self.assertGreater(channels_wgs84.id.size, 0)
        self.assertEqual(len(feat), channels_wgs84.id.size)
        self.assertEqual(feat[0]["properties"]["object_type"], "v2_channel")

    def test_weir_serializer(self):
        weirs = self.parser.lines.weirs.subset("1D_ALL").reproject_to("4326")
        wjson = WeirsGeoJsonSerializer(weirs).data
        j = json.loads(wjson)
        feat = j["features"]
        self.assertGreater(weirs.id.size, 0)
        self.assertEqual(len(feat), weirs.id.size)
        self.assertEqual(feat[0]["properties"]["object_type"], "v2_weir")

    def test_breach_serializer(self):
        breaches = self.parser.breaches.filter(kcu=55).reproject_to("4326")
        bjson = BreachesGeoJsonSerializer(breaches).data
        j = json.loads(bjson)
        feat = j["features"]
        self.assertGreater(breaches.id.size, 0)
        self.assertEqual(len(feat), breaches.id.size)
        self.assertEqual(feat[0]["properties"]["object_type"], "v2_breach")

    def test_manhole_serializer(self):
        manholes = self.parser.nodes.manholes.subset("1D_ALL").reproject_to(
            "4326"
        )
        mjson = ManholesGeoJsonSerializer(manholes).data
        j = json.loads(mjson)
        feat = j["features"]
        self.assertGreater(manholes.id.size, 0)
        self.assertEqual(len(feat), manholes.id.size)
        self.assertEqual(
            feat[0]["properties"]["object_type"], "v2_connection_nodes"
        )

    def test_pump_serializer(self):
        pumps = self.parser.pumps.reproject_to("4326")
        pjson = PumpsGeoJsonSerializer(pumps).data
        j = json.loads(pjson)
        feat = j["features"]
        self.assertGreater(pumps.id.size, 0)
        self.assertEqual(len(feat), pumps.id.size)
        self.assertEqual(
            feat[0]["properties"]["object_type"], "v2_pumpstation"
        )



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
import pytest

from threedigrid.admin.gridresultadmin import GridH5ResultAdmin
from threedigrid.admin.nodes.models import Nodes
from threedigrid.orm.base.fields import TimeSeriesCompositeArrayField

from threedigrid.orm.base.options import Options

test_file_dir = os.path.join(os.getcwd(), "tests/test_files")

# the testfile is a copy of the v2_bergermeer gridadmin file
result_file = os.path.join(test_file_dir, "results_3di.nc")
grid_file = os.path.join(test_file_dir, "gridadmin.h5")


@pytest.fixture()
def gr():
    gr = GridH5ResultAdmin(grid_file, result_file)
    yield gr
    gr.close()


@pytest.fixture()
def opt_nodes(gr):
    opt = Options(gr.nodes)
    yield opt


@pytest.fixture()
def opt_lines(gr):
    opt = Options(gr.lines)
    yield opt


@pytest.fixture()
def opt_breaches(gr):
    opt = Options(gr.breaches)
    yield opt


@pytest.fixture()
def opt_pumps(gr):
    opt = Options(gr.pumps)
    yield opt


def test_options_nodes(opt_nodes):
    assert isinstance(opt_nodes.inst, Nodes)


def test_options_get_field(opt_lines):
    field = opt_lines.get_field("au")
    assert isinstance(field, TimeSeriesCompositeArrayField)


def test_breach_options_get_fields(opt_breaches):
    fields = opt_breaches.get_fields()
    assert set(fields.keys()) == {
        "levbr",
        "breach_width",
        "levl",
        "breach_depth",
        "coordinates",
        "kcu",
        "levmat",
        "seq_ids",
        "content_pk",
        "id",
    }


def test_pump_options_get_fields_only_names(opt_pumps):
    fields = opt_pumps.get_fields(only_names=True)

    assert set(fields) == {
        "zoom_category",
        "content_pk",
        "type",
        "bottom_level",
        "display_name",
        "lower_stop_level",
        "node_coordinates",
        "coordinates",
        "start_level",
        "q_pump",
        "capacity",
        "id",
        "node2_id",
        "node1_id",
    }


def test_nodes_options_get_composite_meta(opt_nodes):
    assert opt_nodes._get_composite_meta("s1", "units") == b"m"


def test_lines_options_get_meta_values(opt_lines):
    v = opt_lines._get_meta_values("q")
    assert set(v.get("q")) == {None, b"Discharge on flow line", b"m3 s-1"}


def test_smoke_looup_idx(opt_nodes):
    idx = opt_nodes._get_lookup_index()
    assert len(idx.shape) == 1 and idx.size > 0



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import pytest
import numpy as np

from threedigrid.admin.nodes.models import Nodes


def test_nodes_timeseries_start_end_time_kwargs(gr):
    ts = gr.nodes.timestamps
    qs_s1 = gr.nodes.timeseries(start_time=ts[0], end_time=ts[6]).s1
    assert qs_s1.shape[0] == ts[0: 6 + 1].size


def test_nodes_timeseries_with_subset(gr):
    qs_s1 = gr.nodes.subset("1d_all").timeseries(start_time=0, end_time=500).s1
    assert qs_s1.shape[0] == 9 and qs_s1.shape[1] > 0


def test_nodes_timeseries_start_time_only_kwarg(gr):
    ts = gr.nodes.timestamps
    qs = gr.nodes.timeseries(start_time=ts[6])
    assert qs.s1.shape[0] == ts[6:].size


def test_nodes_timeseries_end_time_only_kwarg(gr):
    ts = gr.nodes.timestamps
    qs = gr.nodes.timeseries(end_time=ts[6])
    assert qs.s1.shape[0] == ts[: 6 + 1].size


def test_nodes_timeseries_index_filter(gr):
    qs_s1 = gr.nodes.timeseries(indexes=[1, 2, 3]).s1
    assert qs_s1.shape[0] == 3


def test_nodes_timeseries_slice_filter(gr):
    qs_s1 = gr.nodes.timeseries(indexes=slice(1, 4)).s1
    assert qs_s1.shape[0] == 3


def test_pump_timeseries_slice_filter(gr):
    qs = gr.pumps.timeseries(indexes=slice(1, 4)).q_pump
    assert qs.shape[0] == 3


def test_pump_timeseries_index_filter(gr):
    qs = gr.pumps.timeseries(indexes=[1, 2, 3]).q_pump
    assert qs.shape[0] == 3


def test_breach_timeseries_slice_filter(gr):
    qs = gr.breaches.timeseries(indexes=slice(1, 4)).breach_depth
    assert qs.shape[0] == 3


def test_breach_timeseries_index_filter(gr):
    qs = gr.breaches.timeseries(indexes=[1, 2, 3]).breach_width
    assert qs.shape[0] == 3


def test_lines_timeseries_index_filter(gr):
    qs_u1 = gr.lines.timeseries(indexes=[1, 2, 3, 4, 5]).u1
    assert qs_u1.shape[0] == 5


def test_lines_timeseries_slice_filter(gr):
    qs_u1 = gr.lines.timeseries(indexes=slice(1, 4)).u1
    assert qs_u1.shape[0] == 3


def test_lines_timeseries_with_subset(gr):
    qs_u1 = gr.lines.subset("1d_all").timeseries(indexes=slice(1, 4)).u1
    assert qs_u1.shape[0] == 3 and qs_u1.shape[1] > 0


def test_set_timeseries_chunk_size(gr):
    # default should be 10
    assert gr.timeseries_chunk_size == 10
    gr.set_timeseries_chunk_size(15)
    assert gr.timeseries_chunk_size == 15


def test_missing_kwargs_raises_key_error(gr):
    # default should be 10
    with pytest.raises(KeyError):
        gr.lines.timeseries().u1


def test_index_key_raises_type_error(gr):
    # default should be 10
    with pytest.raises(TypeError):
        gr.lines.timeseries(indexes="wrong").u1


def test_set_timeseries_chunk_size_raises_value_error(gr):
    # default should be 10
    with pytest.raises(ValueError):
        gr.set_timeseries_chunk_size(0)

    with pytest.raises(ValueError):
        gr.set_timeseries_chunk_size(-5)

    with pytest.raises(ValueError):
        gr.set_timeseries_chunk_size("20.5")


def test_get_timeseries_mask_filter(gr):
    assert gr.timeseries_chunk_size == 10
    gr.set_timeseries_chunk_size(50)
    mask_f = gr.nodes.get_timeseries_mask_filter()
    assert mask_f == slice(0, 50, None)


def test_timestamps(gr):
    np.testing.assert_array_equal(gr.nodes.timestamps, gr.lines.timestamps)
    n_qs = gr.nodes.timeseries(start_time=0, end_time=500)
    l_qs = gr.lines.timeseries(start_time=0, end_time=500)
    np.testing.assert_array_equal(n_qs.timestamps, l_qs.timestamps)


def test_dt_timestamps(gr):
    n_qs = gr.nodes.timeseries(start_time=0, end_time=500)
    assert len(n_qs.dt_timestamps) == len(n_qs.timestamps)


def test_dt_timeseries_sampling(gr):
    s1 = gr.nodes.timeseries(start_time=0).sample(5).s1
    assert len(s1) == 5
    s1 = gr.nodes.timeseries(start_time=1).sample(5).s1
    assert len(s1) == 5
    s1 = gr.nodes.timeseries(indexes=slice(1, None)).sample(5).s1
    assert len(s1) == 5


def test_get_model_instance_by_field_name(gr):
    inst = gr.get_model_instance_by_field_name("s1")
    assert isinstance(inst, Nodes)


def test_get_model_instance_by_field_name_raises_index_error_unknown_field(gr):
    with pytest.raises(IndexError):
        gr.get_model_instance_by_field_name("unknown_field")


def test_get_model_instance_by_field_name_raises_index_error(gr):
    with pytest.raises(IndexError):
        gr.get_model_instance_by_field_name("zoom_category")


def test_gr_get_subset_idx_leak(gr):
    assert gr.nodes._get_subset_idx('leak').shape == (10748,)


def test_gr_get_subset_ids_no_composite_field(gr):
    assert gr.nodes._get_subset_idx('s1') is None


def test_gr_chain_filter(gr):
    assert gr.nodes.filter(
        id=4).get_filtered_field_value('leak').shape == (9, 1)

# commented for now until the new aggregate.nc is finished

# def test_get_node_aggregate_netcdf_results(agg_gr):
#     assert 's1_max' in agg_gr.netcdf_file.variables.keys()
#     assert 'vol_max' in agg_gr.netcdf_file.variables.keys()
#     assert hasattr(agg_gr.nodes, 's1_max')
#     assert hasattr(agg_gr.nodes, 'vol_max')
#     assert agg_gr.nodes.s1_max.shape[0] > 0
#     assert agg_gr.nodes.vol_max.shape[0] > 0
#
#
# def test_get_line_aggregate_netcdf_results(agg_gr):
#     assert 'q_max' in agg_gr.netcdf_file.variables.keys()
#     assert 'u1_max' in agg_gr.netcdf_file.variables.keys()
#     assert hasattr(agg_gr.lines, 'q_max')
#     assert hasattr(agg_gr.lines, 'u1_max')
#     assert agg_gr.lines.q_max.shape[0] > 0
#     assert agg_gr.lines.u1_max.shape[0] > 0


# def test_get_node_netcdf_results(gr):
#     assert 's1' in gr.netcdf_file.variables.keys()
#     assert 'vol' in gr.netcdf_file.variables.keys()
#     assert hasattr(gr.nodes, 's1')
#     assert hasattr(gr.nodes, 'vol')
#     assert gr.nodes.s1.shape[0] > 0
#     assert gr.nodes.vol.shape[0] > 0



import h5py


class H5SwmrFile(h5py.File):
    """
    Override h5py file access, to
    automatically refresh datasets and make
    sure we only have one reference to one dataset
    """

    def __init__(self, path, file_modus):
        self._datasets = {}
        super(H5SwmrFile, self).__init__(
            path, file_modus, swmr=True)

        # By default register all datasets
        for key in self.keys():
            self.get(key)

    def get(self, key):
        return self.__getitem__(key)

    def refresh_datasets(self):
        for _, dataset in self._datasets.items():
            dataset.refresh()

    def __getitem__(self, key):
        """
        Store datasets pointers in owned dictionairy
        """
        if key in self._datasets:
            return self._datasets[key]

        res = super(H5SwmrFile, self).__getitem__(key)
        if isinstance(res, h5py._hl.dataset.Dataset):
            if key not in self._datasets:
                self._datasets[key] = res
        return res



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
import json
from threedigrid.admin.pumps.models import Pumps
from threedigrid.admin import constants
from threedigrid.geo_utils import raise_import_exception
from threedigrid.orm.base.encoder import NumpyEncoder
from six.moves import range

try:
    import geojson
except ImportError:
    geojson = None


def format_to_str(value, empty_value='--', format_string='%0.2f'):
    """
    Format value with format_string when evaluated as True
    else return the empty value
    """
    if not value:
        return empty_value
    return format_string % value


class PumpsGeoJsonSerializer():
    def __init__(self, pumps=None, data=None, indent=None):
        if geojson is None:
            raise_import_exception('geojson')
        if pumps:
            assert isinstance(pumps, Pumps)
        self._data = data
        self._pumps = pumps
        self._indent = indent

    @property
    def geos(self):
        if self._pumps:
            selection = self._pumps.to_dict()
        else:
            selection = self._data

        geos = []
        if 'coordinates' not in selection:
            raise ValueError(
                "Can't return data as geojson "
                "for selection without geometries")

        for i in range(selection['id'].shape[-1]):
            pt = geojson.Point(
                [round(x, constants.LONLAT_DIGITS)
                 for x in selection['coordinates'][:, i]])
            cn_meta = [
                    ['object_type', constants.TYPE_V2_PUMPSTATION],
                    ['display name', selection['display_name'][i]],
                    ['idx', str(selection['id'][i])],
                    ['start level', '%s [m MSL]' % format_to_str(
                        selection['start_level'][i])],
                    ['lower stop level', '%s [m MSL]' % format_to_str(
                        selection['lower_stop_level'][i])],
                    ['capacity', '%s [L/s]' % format_to_str(
                        selection['capacity'][i] * 1000.0
                        if selection['capacity'][i] else False)],
                    ['bottom level', '%s [m MSL]' % format_to_str(
                        selection['bottom_level'][i])],
                    ['nod1d_a', selection['node1_id'][i]],
                    ['nod1d_b', selection['node2_id'][i]
                        if selection['node2_id'][i] != -9999 else '--'],
                    ]
            cn_props = dict(
                object_type=constants.TYPE_V2_PUMPSTATION,
                meta=cn_meta,
                pump_idx=int(selection['id'][i]),
                )
            feat = geojson.Feature(
                geometry=pt,
                properties=cn_props,
                )
            geos.append(feat)

        return geos

    @property
    def data(self):
        geos = self.geos
        return json.dumps({
            'type': 'FeatureCollection',
            'features': geos,
        }, indent=self._indent, cls=NumpyEncoder)



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
import numpy as np
from . import constants
import logging

from threedigrid.admin.utils import get_or_create_group
from threedigrid.admin.constants import TYPE_CODE_MAP
import six

logger = logging.getLogger(__name__)


def get_id_map(objs_for_id_map, breach_dict):
    if not objs_for_id_map:
        return

    id_map = {}

    class_map = {TYPE_CODE_MAP[arg[0].__tablename__]: arg
                 for arg in objs_for_id_map if arg}
    for name, collection in six.iteritems(class_map):
        collection_map = {
            entry.pk: abs(entry.id) for entry in collection
        }
        id_map[name] = collection_map

    if breach_dict:
        id_map[TYPE_CODE_MAP['v2_breach']] = breach_dict
    return id_map


class IdMapper(object):

    LINE_TYPES = [
        constants.TYPE_V2_PIPE, constants.TYPE_V2_CHANNEL,
        constants.TYPE_V2_CULVERT,
        constants.TYPE_V2_ORIFICE, constants.TYPE_V2_WEIR]

    def __init__(self, id_mapping):
        self._id_mapping = id_mapping
        self.obj_slices = None
        self._define_id_obj_slices()

    @property
    def id_mapping(self):
        return self._id_mapping.value

    def get_by_name(self, obj_name):
        """
        get a slice of ``obj_slices``. That is, a structured array for
        one specific object type (like v2_channels)

        :param obj_name: name of a given object, for example v2_channel

        :return: a structured array with fields
            'obj_code', '<i4',
            ('pk', '<i4'),
            ('seq_id', '<i4')
        """
        obj_code = constants.TYPE_CODE_MAP[obj_name]
        idx = self.obj_slices[obj_code]
        return self.id_mapping[idx]

    def get_by_code(self, obj_code):
        """
        get a slice of ``obj_slices``. That is, a structured array for
        one specific object type (like v2_channels)

        :param obj_code: code for a given object, for example 2 for
            v2_channel. See TYPE_CODE_MAP in threedigrid.admin.constants
        :return: a structured array with fields
            'obj_code', '<i4',
            ('pk', '<i4'),
            ('seq_id', '<i4')

        """
        idx = self.obj_slices[obj_code]
        return self.id_mapping[idx]

    def get_pk(self, obj_name, seq_id):
        obj_code = constants.TYPE_CODE_MAP[obj_name]
        sl = self.obj_slices[obj_code]
        return self.id_mapping[sl]['pk'][np.where(
            self.id_mapping[sl]['seq_id'] == seq_id)][0]

    def _define_id_obj_slices(self):
        obj_codes = np.unique(self.id_mapping['obj_code'])
        self.obj_slices = {
            n: np.where(self.id_mapping['obj_code'] == n)
            for n in obj_codes
        }

    @staticmethod
    def prepare_mapper(h5py_file, threedi_datasource):
        id_map = get_id_map(
            threedi_datasource.objs_for_id_map,
            threedi_datasource.breach_dict)

        def _get_id_mapping_entry():
            for k, v in six.iteritems(id_map):
                for pk, seq_id in six.iteritems(v):
                    yield k, pk, seq_id

        dtype_dict = {'formats': [
                'i4', 'i4', 'i4'],
                'names': ['obj_code', 'pk', 'seq_id']
        }
        id_mapping = np.fromiter(
            _get_id_mapping_entry(),
            dtype=dtype_dict,
            count=sum(len(v) for v in six.itervalues(id_map))
        )
        try:
            gr = get_or_create_group(
                h5py_file, constants.GROUP_MAPPINGS
            )
            dset = gr.get(constants.DSET_ID_MAPPING, None)
            if not dset:
                gr.create_dataset(
                    constants.DSET_ID_MAPPING,
                    id_mapping.shape,
                    dtype=dtype_dict, data=id_mapping
                )
            logger.info(
                '[+] Successfully added {} id mapping data to grid admin file'
            )
        except Exception:
            logger.exception('Failed to add levee to hdf5')
            raise



from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
import tempfile

from click.testing import CliRunner

from threedigrid.management import help_texts
from threedigrid.management.commands.kick import kick_start
from threedigrid.management.commands.kick import export_to


test_file_dir = os.path.join(os.getcwd(), "tests/test_files")

# the testfile is a copy of the v2_bergermeer gridadmin file
grid_admin_h5_file = os.path.join(test_file_dir, "gridadmin.h5")


def test_model_overview():
    # smoke test
    txt = help_texts.model_overview(grid_admin_h5_file)
    assert txt != ""


def test_command_kick_start():
    # smoke test
    runner = CliRunner()
    result = runner.invoke(
        kick_start, [u"--grid-file={}".format(grid_admin_h5_file), u"--no-ipy"]
    )
    assert result.exit_code == 0


def test_command_export_to():
    # smoke test

    d = tempfile.mkdtemp()
    test_shp = os.path.join(d, "cmd_exporter_test_lines.shp")

    runner = CliRunner()
    result = runner.invoke(
        export_to,
        [
            u"--grid-file={}".format(grid_admin_h5_file),
            u"--file-type=shape",
            u"--output-file={}".format(test_shp),
            u"--model=lines",
            u"--subset=1d_all",
        ],
    )
    assert result.exit_code == 0
    assert os.path.exists(test_shp)



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
import pytest

import numpy as np


from threedigrid.admin.gridresultadmin import GridH5AggregateResultAdmin
from threedigrid.admin.lines.models import Lines
from threedigrid.admin.nodes.models import Nodes


test_file_dir = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "test_files"
)

# the testfile is a copy of the v2_bergermeer gridadmin file
result_file = os.path.join(test_file_dir, "results_3di.nc")
agg_result_file = os.path.join(test_file_dir, "aggregate_results_3di.nc")
grid_file = os.path.join(test_file_dir, "gridadmin.h5")


@pytest.fixture()
def agg_gr():
    gr = GridH5AggregateResultAdmin(grid_file, agg_result_file)
    yield gr
    gr.close()


def test_get_meta_composite_fields_nodes(agg_gr):
    assert set(agg_gr.nodes.Meta.composite_fields.keys()) == {
        "_mesh_id",
        "q_lat_avg",
        "q_lat_cum",
        "q_lat_cum_negative",
        "q_lat_cum_positive",
        "q_lat_max",
        "q_lat_min",
        "rain_avg",
        "rain_cum",
        "rain_cum_negative",
        "rain_cum_positive",
        "rain_max",
        "rain_min",
        "s1_avg",
        "s1_max",
        "s1_min",
        "su_avg",
        "su_max",
        "su_min",
        "vol_avg",
        "vol_current",
        "vol_max",
        "vol_min",
        "vol_sum",
    }


def test_get_meta_subset_fields_nodes(agg_gr):
    assert set(agg_gr.nodes.Meta.subset_fields.keys()) == {
        "infiltration_rate_simple_avg",
        "infiltration_rate_simple_cum",
        "infiltration_rate_simple_cum_negative",
        "infiltration_rate_simple_cum_positive",
        "infiltration_rate_simple_max",
        "infiltration_rate_simple_min",
        "intercepted_volume_avg",
        "intercepted_volume_current",
        "intercepted_volume_max",
        "intercepted_volume_min",
        "intercepted_volume_cum",
        "leak_avg",
        "leak_cum",
        "leak_cum_negative",
        "leak_cum_positive",
        "leak_max",
        "leak_min",
        "ucx_avg",
        "ucx_max",
        "ucx_min",
        "ucy_avg",
        "ucy_max",
        "ucy_min",
        "q_sss_avg",
        "q_sss_cum",
        "q_sss_cum_negative",
        "q_sss_cum_positive",
        "q_sss_max",
        "q_sss_min",
    }


def test_get_meta_fields_nodes(agg_gr):
    assert set(agg_gr.nodes._meta.get_fields(only_names=True)) == {
        "cell_coords",
        "content_pk",
        "coordinates",
        "id",
        "infiltration_rate_simple_cum",
        "intercepted_volume_cum",
        "is_manhole",
        "leak_cum",
        "node_type",
        "q_lat_cum",
        "rain_avg",
        "rain_cum",
        "s1_max",
        "s1_min",
        "seq_id",
        "su_min",
        "vol_current",
        "vol_max",
        "vol_min",
        "zoom_category",
        "sumax",
    }


def test_get_timestamps_nodes(agg_gr):
    ts = agg_gr.nodes.get_timestamps("rain_avg")
    assert isinstance(ts, np.ndarray)
    assert ts.size > 0


def test_get_timestamps_raises_attribute_error(agg_gr):
    with pytest.raises(AttributeError):
        agg_gr.nodes.get_timestamps("does_not_exist")


def test_get_timestamps_lines(agg_gr):
    ts = agg_gr.nodes.get_timestamps("q_cum")
    assert isinstance(ts, np.ndarray)
    assert ts.size > 0


def test_get_time_unit(agg_gr):
    tu = agg_gr.nodes.get_time_unit("q_cum")
    assert tu.startswith(b"seconds since ")


def test_get_timestamps_pumps(agg_gr):
    ts = agg_gr.nodes.get_timestamps("q_cum")
    assert isinstance(ts, np.ndarray)
    assert ts.size > 0


def test_nodes_timeseries_start_end_time_kwargs(agg_gr):
    ts = agg_gr.nodes.get_timestamps("s1_max")
    qs_s1_max = agg_gr.nodes.timeseries(
        start_time=ts[0], end_time=ts[1]
    ).s1_max
    assert qs_s1_max.shape[0] == 2


def test_nodes_timeseries_with_subset(agg_gr):
    ts = agg_gr.nodes.get_timestamps("s1_max")
    qs_s1_max = (
        agg_gr.nodes.subset("1d_all")
        .timeseries(start_time=ts[0], end_time=ts[1])
        .s1_max
    )
    assert qs_s1_max.shape[0] == 2 and qs_s1_max.shape[1] > 0


def test_nodes_timeseries_start_time_only_kwarg(agg_gr):
    ts = agg_gr.nodes.get_timestamps("s1_max")
    qs_s1_max = agg_gr.nodes.timeseries(start_time=ts[1]).s1_max
    assert qs_s1_max.shape[0] == 6


def test_nodes_timeseries_end_time_only_kwarg(agg_gr):
    ts = agg_gr.nodes.get_timestamps("s1_max")
    qs_s1_max = agg_gr.nodes.timeseries(end_time=ts[1])
    assert qs_s1_max.s1_max.shape[0] == 2


def test_nodes_timeseries_slice_filter(agg_gr):
    qs_s1_min = agg_gr.nodes.timeseries(indexes=slice(0, 1)).s1_min
    assert qs_s1_min.shape[0] == 1


def test_get_model_instance_by_agg_field_name(agg_gr):
    inst = agg_gr.get_model_instance_by_field_name("rain_avg")
    assert isinstance(inst, Nodes)


def test_get_model_instance_by_agg_field_name_lines(agg_gr):
    inst = agg_gr.get_model_instance_by_field_name("q_cum_positive")
    assert isinstance(inst, Lines)


def test_reading_grid_result_and_grid_aggregate_result(gr, agg_gr):
    assert hasattr(gr.nodes, "s1")
    assert not hasattr(gr.nodes, "rain_avg")
    assert hasattr(agg_gr.nodes, "rain_avg")
    assert not hasattr(agg_gr.nodes, "s1")


# TODO once the threedicore has implemented aggregations for pumpstations
# TODO and breaches also implement some tests
# def test_pump_timeseries_slice_filter(agg_gr):
#     qs = agg_gr.pumps.timeseries(indexes=slice(1, 4)).q_pump
#     assert qs.shape[0] == 3
#
#
# def test_pump_timeseries_index_filter(agg_gr):
#     qs = agg_gr.pumps.timeseries(indexes=[1, 2, 3]).q_pump
#     assert qs.shape[0] == 3
#
#
# def test_breach_timeseries_slice_filter(agg_gr):
#     qs = agg_gr.breaches.timeseries(indexes=slice(1, 4)).depth
#     assert qs.shape[0] == 3
#
#
# def test_breach_timeseries_index_filter(agg_gr):
#     qs = agg_gr.breaches.timeseries(indexes=[1, 2, 3]).width
#     assert qs.shape[0] == 3
#
#
# def test_lines_timeseries_index_filter(agg_gr):
#     qs_u1 = agg_gr.lines.timeseries(indexes=[1, 2, 3, 4, 5]).u1
#     assert qs_u1.shape[0] == 5
#
# def test_lines_timeseries_slice_filter(agg_gr):
#     qs_u1 = agg_gr.lines.timeseries(indexes=slice(1, 4)).u1
#     assert qs_u1.shape[0] == 3
#
# def test_lines_timeseries_with_subset(agg_gr):
#     qs_u1 = agg_gr.lines.subset('1d_all').timeseries(indexes=slice(1, 4)).u1
#     assert qs_u1.shape[0] == 3 and qs_u1.shape[1] > 0
